local mathMax = math.max;
local mathMin = math.min;
local mathFloor = math.floor;
local mathCeil = math.ceil;

local PADDING = 5;

local DEFAULTICON = "Interface\\Icons\\INV_Misc_QuestionMark";

local COLOR_DEFAULT = {1, 1, 1};
local COLOR_FG = {1, 0.5, 0};
local COLOR_FG_DISABLE = {0.4, 0.4, 0.4};
local COLOR_BG = {0.1, 0.1, 0.1, 1};
local COLOR_TEXT = {0.8, 0.8, 0.8};
local COLOR_TEXT_DISABLE = {0.4, 0.4, 0.4};

local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

local function dummy()
	return;
end

----------------------------------------------------------------	MessButtonTemplate
local MessButtonMeta = {
	["enableColor"] = COLOR_FG,
	["disableColor"] = COLOR_FG_DISABLE,
	["Update"] = dummy,
};

function MessButton_OnLoad(self)
	inherits(self, MessButtonMeta);
end

function MessButton_OnEnter(self, motion)
	local anim = self:IsEnabled() and self.highlightIn or self.highlightInAlter;
	self.highlight:Show();
	anim:Play();
end

function MessButton_OnLeave(self, motion)
	self.highlightIn:Stop();
	self.highlightInAlter:Stop();
	self.highlight:Hide();
end

function MessButton_Update(self)
	self:Update();
end

----------------------------------------------------------------	MessTextButtonTemplate
local MessTextButtonMeta = {
	["enableColor"] = COLOR_TEXT,
	["disableColor"] = COLOR_TEXT_DISABLE,
	["SetText"] = function(self, text, noResize)
		self.text:SetText(text or "MessTextButton");
		if not noResize then
			self:SetHeight(self.text:GetStringHeight() + PADDING * 2);
			self:SetWidth(self.text:GetStringWidth() + PADDING * 2);
		end
		self:Update();
	end,
	["Update"] = function(self)
		if self:IsEnabled() then
			self.text:SetTextColor(unpack(self.enableColor));
		else
			self.text:SetTextColor(unpack(self.disableColor));
		end
	end,
};

function MessTextButton_OnLoad(self)
	MessButton_OnLoad(self);
	inherits(self, MessTextButtonMeta);
	
	self:Update();
end

----------------------------------------------------------------	MessIconButtonTemplate
local MessIconButtonMeta = {
	["SetIcon"] = function(self, texture)
		self.icon:SetTexture(texture or DEFAULTICON);
		self:Update();
	end,
	["GetIcon"] = function(self)
		return self.icon:GetTexture();
	end,
	["Update"] = function(self)
		if self:IsEnabled() then
			self.icon:SetDesaturated(false);
			self.LBorder:SetVertexColor(unpack(self.enableColor));
			self.RBorder:SetVertexColor(unpack(self.enableColor));
			self.TBorder:SetVertexColor(unpack(self.enableColor));
			self.BBorder:SetVertexColor(unpack(self.enableColor));
		else
			self.icon:SetDesaturated(true);
			self.LBorder:SetVertexColor(unpack(self.disableColor));
			self.RBorder:SetVertexColor(unpack(self.disableColor));
			self.TBorder:SetVertexColor(unpack(self.disableColor));
			self.BBorder:SetVertexColor(unpack(self.disableColor));
		end
	end,
};

function MessIconButton_OnLoad(self)
	MessButton_OnLoad(self);
	inherits(self, MessIconButtonMeta);
	
	self:Update();
end

----------------------------------------------------------------	MessCleanIconButtonTemplate
local MessCleanIconButtonMeta = {
	["enableColor"] = COLOR_TEXT,
	["disableColor"] = COLOR_TEXT_DISABLE,
	["SetIcon"] = function(self, texture)
		texture = texture or DEFAULTICON;
		self.icon:SetTexture(texture);
		self.highlight:SetTexture(texture);
		self:Update();
	end,
	["GetIcon"] = function(self)
		return self.icon:GetTexture();
	end,
	["Update"] = function(self)
		if self:IsEnabled() then
			self.icon:SetVertexColor(unpack(self.enableColor));
		else
			self.icon:SetVertexColor(unpack(self.disableColor));
		end
	end,
};

function MessCleanIconButton_OnLoad(self)
	MessButton_OnLoad(self);
	inherits(self, MessCleanIconButtonMeta);
	
	self:Update();
end

----------------------------------------------------------------	MessHybridButtonTemplate
local MessHybridButtonMeta = {
	["enableColor"] = COLOR_TEXT,
	["disableColor"] = COLOR_TEXT_DISABLE,
	["SetIcon"] = function(self, texture)
		self.icon:SetTexture(texture or DEFAULTICON);
		self:Update();
	end,
	["GetIcon"] = function(self)
		return self.icon:GetTexture();
	end,
	["SetText"] = function(self, text)
		self.text:SetText(text or "");
		self:Update();
	end,
	["GetText"] = function(self)
		return self.text:GetText();
	end,
	["Update"] = function(self)
		if self:IsEnabled() then
			self.icon:SetVertexColor(unpack(self.enableColor));
			self.text:SetTextColor(unpack(self.enableColor));
		else
			self.icon:SetVertexColor(unpack(self.disableColor));
			self.text:SetTextColor(unpack(self.disableColor));
		end
	end,
};

function MessHybridButton_OnLoad(self)
	MessButton_OnLoad(self);
	inherits(self, MessHybridButtonMeta);
	
	self:Update();
end

----------------------------------------------------------------	MessButtonGroupButtonTemplate
local MessButtonGroupButtonMeta = {
	["Update"] = function(self)
		local parent = self:GetParent();
		if self:IsEnabled() then
			self.icon:SetVertexColor(unpack(parent.enableColor));
		else
			self.icon:SetVertexColor(unpack(parent.disableColor));
		end
	end,
};

function MessButtonGroupButton_OnLoad(self)
	MessCleanIconButton_OnLoad(self);
	inherits(self, MessButtonGroupButtonMeta);
	self:Update();
end

function MessButtonGroupButton_OnEnter(self, button)
	local anim = self:IsEnabled() and self.highlightIn or self.highlightInAlter;
	self.highlight:Show();
	anim:Play();
	local parent = self:GetParent();
	parent:SetTip(self.tipText);
end

function MessButtonGroupButton_OnLeave(self, button)
	self.highlightIn:Stop();
	self.highlightInAlter:Stop();
	self.highlight:Hide();
	local parent = self:GetParent();
	parent:SetTip();
end

function MessFloatButtonGroupButton_PostClick(self, ...)	--	triggers parentOfParent update
	local delegated = self:GetParent():GetParent();
	if delegated then
		if delegated:IsMouseOver() then
			pcall(delegated:GetScript("OnEnter"), delegated, ...);
		else
			pcall(delegated:GetScript("OnLeave"), delegated, ...);
		end
	end
end

----------------------------------------------------------------	MessButtonGroupSeparatorTemplate
local MessButtonGroupSeparatorMeta = {
	["Update"] = function(self)
		local parent = self:GetParent();
		self.icon:SetVertexColor(unpack(parent.enableColor));
	end,
};

function MessButtonGroupSeparator_OnLoad(self)
	inherits(self, MessButtonGroupSeparatorMeta);
	self.icon:SetTexCoord(1, 0, 0, 0, 1, 1, 0, 1);
	self.isSeparator = true;
	self:Update();
end

----------------------------------------------------------------	MessFloatButtonGroupTemplate
local MessFloatButtonGroupMeta = {
	--	attributes
	["_items"] = nil,
	["_pool"] = nil,
	["_count"] = 0,
	["enableColor"] = COLOR_TEXT,
	["disableColor"] = COLOR_TEXT_DISABLE,
	["tipTextColor"] = COLOR_TEXT,
	--	methods
	["_redistribute"] = function(self)
		local num = #self._items;
		if num == 0 then return; end
		local pivot, totalWidth = self._items[num], self._items[num]:GetWidth();
		for index = num - 1, 1, -1 do
			local item = self._items[index];
			item:ClearAllPoints();
			item:SetPoint("RIGHT", pivot, -totalWidth - PADDING, 0);
			totalWidth = totalWidth + item:GetWidth() + PADDING;
		end
		pivot:SetPoint("CENTER", totalWidth / 2, 0);
		self.text:SetPoint("LEFT", pivot, "RIGHT", PADDING, 0);
		
		self.modified = nil;
	end,
	["AddButton"] = function(self, texture, tip, func, index, noAdjust)
		local new = tremove(self._pool);
		if not new then
			self._count = self._count + 1;
			new = CreateFrame("Button", "$parentBtn"..self._count, self, "MessButtonGroupButtonTemplate");
		end
		
		if index then
			tinsert(self._items, index, new);
		else
			tinsert(self._items, new);
		end
		self.modified = true;
		
		new:SetIcon(texture);
		if type(tip) == "function" then
			new:SetScript("OnClick", tip);
		else
			new.tipText = tip;
			new:SetScript("OnClick", func);
		end
		new:Show();
		
		if not noAdjust then
			self:_redistribute();
		end
		
		return new;
	end,
	["AddSeparator"] = function(self, index, noAdjust)
		self._count = self._count + 1;
		local new = CreateFrame("Frame", "$parentSep"..self._count, self, "MessButtonGroupSeparatorTemplate");
		
		if index then
			tinsert(self._items, index, new);
		else
			tinsert(self._items, new);
		end
		self.modified = true;
		
		if not noAdjust then
			self:_redistribute();
		end
	end,
	["RemoveButton"] = function(self, index, noAdjust)
		local btn = tremove(self._items, index);
		if btn then
			self.modified = true;
			btn:ClearAllPoints();
			btn:Hide();
			if not btn.isSeparator then
				tinsert(self._pool, btn);
			end
			if not noAdjust then
				self:_redistribute();
			end
		end
	end,
	["ClearAll"] = function(self)
		for _, item in ipairs(self._items) do
			item:ClearAllPoints();
			item:Hide();
			if not item.isSeparator then
				tinsert(self._pool, item);
			end
		end
		wipe(self._items);
		self.modified = true;
	end,
	["GetButton"] = function(self, index)
		return self._items[index];
	end,
	["GetButtonNum"] = function(self)
		return #self._items;
	end,
	["SetTip"] = function(self, text)
		self.text:SetText(text);
		self.text:SetTextColor(unpack(self.tipTextColor));
		if text then
			self.textFadein:Play();
		else
			self.textFadein:Stop();
		end
	end,
	["Update"] = function(self)
		for _, item in ipairs(self._items) do
			item:Update();
			if not item.isSeparator then
				pcall(item:GetScript(item:IsMouseOver() and "OnEnter" or "OnLeave"), item);
			end
		end
		if self.modified then
			self:_redistribute();
		end
	end,
	["AttachTo"] = function(self, frame)
		self:SetParent(frame);
		self:ClearAllPoints();
		self:SetAllPoints(frame);
		self:Show();
		self:Raise();
	end,
};

function MessFloatButtonGroup_OnLoad(self)
	inherits(self, MessFloatButtonGroupMeta);
	self._items = {};
	self._pool = {};
end

----------------------------------------------------------------	MessSizerTemplate
function MessSizer_OnLoad(self)
	self:GetNormalTexture():SetDesaturated(true);
	self:RegisterForDrag("LeftButton");
end

function MessSizer_OnDragStart(self, button)
	self:LockHighlight();
	self:GetParent():StartSizing("BOTTOMRIGHT");
end

function MessSizer_OnDragStop(self)
	self:UnlockHighlight();
	local parent = self:GetParent();
	pcall(parent.onResizedFunc, self);
	parent:StopMovingOrSizing();
end

----------------------------------------------------------------	MessCloserTemplate
function MessCloser_OnClick(self, button)
	local parent = self:GetParent();
	pcall(parent.onClosedFunc, self);
	parent:Hide();
end

----------------------------------------------------------------	MessMoverTemplate
function MessMover_OnLoad(self)
	self:RegisterForDrag("LeftButton");
	
	self.bg:SetVertexColor(unpack(COLOR_FG));
	self.text:SetTextColor(unpack(COLOR_BG));
end

function MessMover_OnDragStart(self, button)
	self:GetParent():StartMoving();
end

function MessMover_OnDragStop(self)
	local parent = self:GetParent();
	pcall(parent.onMovedFunc, self);
	parent:StopMovingOrSizing();
end

