local mathMax = math.max;
local mathMin = math.min;
local mathFloor = math.floor;
local mathCeil = math.ceil;

local COLOR_FG = {1, 0.5, 0, 1};
local COLOR_BG = {0.1, 0.1, 0.1, 1};

local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

local function dummy()
	return;
end

local function setVisible(region, visible)
	if visible then
		region:Show();
	else
		region:Hide();
	end
end

----------------------------------------------------------------	MessContainerFrameTemplate
local MessContainerMeta = {
	["_hookedKeyFuncMap"] = nil,
	["_moverDisabled"] = false,
	["_closerDisabled"] = false,
	["_sizerDisabled"] = false,
	["_wheelBlockerDisabled"] = false,
	--	functions
	["_keyInputFilter"] = function(self, key)
		self:SetPropagateKeyboardInput(false);
		if self._hookedKeyFuncMap[key] then
			pcall(self._hookedKeyFuncMap[key], self, key);
		else
			self:SetPropagateKeyboardInput(true);
		end
	end,
	["SetBorderColor"] = function(self, r, g, b, a)
		self.LBorder:SetVertexColor(r,g,b,a);
		self.RBorder:SetVertexColor(r,g,b,a);
		self.BBorder:SetVertexColor(r,g,b,a);
		self.TBorder:SetVertexColor(r,g,b,a);
		self.header.bg:SetVertexColor(r,g,b,a);
	end,
	["SetTitle"] = function(self, text, color)
		if text then
			self.header.text:SetText(text);
		end
		if color then
			self.header.text:SetTextColor(unpack(color));
		end
	end,
	["SetLock"] = function(self, lock)
		self._isLocked = lock;
		setVisible(self.glasspane, lock);
		self:EnableKeyboard(not lock and self._hookedKeyFuncMap);
	end,
	["IsLocked"] = function(self, lock)
		return self.glasspane:IsShown();
	end,
	["EnableMover"] = function(self, enabled)
		self._moverDisabled = not enabled;
		if enabled then
			self.header:Show();
			if not self._closerDisabled then
				self.hcloser:Show();
				self.closer:Hide();
			end
		else
			self.header:Hide();
			if not self._closerDisabled then
				self.hcloser:Hide();
				self.closer:Show();
			end
		end
	end,
	["EnableCloser"] = function(self, enabled)
		self._closerDisabled = not enabled;
		if enabled then
			if not self._moverDisabled then
				self.hcloser:Show();
				self.closer:Hide();
			else
				self.hcloser:Hide();
				self.closer:Show();
			end
		else
			self.hcloser:Hide();
			self.closer:Hide();
		end
	end,
	["EnableSizer"] = function(self, enabled)
		self._sizerDisabled = not enabled;
		setVisible(self.sizer, enabled);
	end,
	["DisableWheelBlocker"] = function(self, disabled)
		self:SetScript("OnMouseWheel", not disabled and dummy);
	end,
	["HookKeyInput"] = function(self, key, func)
		if not self._hookedKeyFuncMap then
			self._hookedKeyFuncMap = {};
			self:EnableKeyboard(true);
			self:SetScript("OnKeyDown", self._keyInputFilter);
		end
		self._hookedKeyFuncMap[key] = func;
	end,
};

function MessContainer_OnLoad(self)
	inherits(self, MessContainerMeta);
	
	self:SetBorderColor(unpack(COLOR_FG));
end

----------------------------------------------------------------	MessMessageFrameTemplate
local MessMessageMeta = {
	["ShowMessage"] = function(self, text)
		local h, w = GetScreenHeight(), GetScreenWidth();
		self:SetHeight(h / 8) ;
		self.text:SetText(text);
		self.text:SetSize(w / 2, h / 16);
		self.tip:SetWidth(w / 2);
		self:Show();
	end,
};

function MessMessageFrame_OnLoad(self)
	inherits(self, MessMessageMeta);
end

function MessMessageFrame_OnClick(self, button)
	self:Hide();
end
