
local Frame_OnEnter = "Frame_OnEnter";
local Frame_OnLeave = "Frame_OnLeave";

local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

----------------------------------------------------------------	MessEventDispatcher
local MessEventDispatcherMeta = {
	["_observers"] = nil,	--	event : (index : observer)
	["_dispatch"] = function(self, event, ...)
		if not self._observers or not self._observers[event] then return; end
		for _, observer in ipairs(self._observers[event]) do
			pcall(observer[event], observer, self, ...);
		end
	end,
	["RegisterCallback"] = function(self, observer, event)
		self._observers = self._observers or {};
		self._observers[event] = self._observers[event] or {};
		for index, registered in ipairs(self._observers[event]) do	--	duplicated test
			if registered == observer then
				return;
			end
		end
		tinsert(self._observers[event], observer);
	end,
	["UnregisterCallback"] = function(self, observer, event)
		if not self._observers or not self._observers[event] then return; end
		for index, registered in ipairs(self._observers[event]) do
			if registered == observer then
				return tremove(self._observers[event], index);
			end
		end
	end,
	["UnregisterAllCallbacks"] = function(self, observer)
		for event, list in pairs(self._observers) do
			for index, registered in ipairs(list) do
				if registered == observer then
					tremove(list, index);
					break;
				end
			end
		end
	end,
};

function MessEventDispatcher_OnLoad(self)
	inherits(self, MessEventDispatcherMeta);
end

----------------------------------------------------------------	MessCursorTracer
local MessCursorTracerMeta = {
	["_traced"] = nil,		--	index : frame
	["_stateMap"] = nil,	--	frame : state
	["SetThreshold"] = function(self, t)
		self.tracer.anim:SetDuration(t);
	end,
	["GetThreshold"] = function(self)
		return self.tracer.anim:GetDuration();
	end,
	["RegisterTracer"] = function(self, frame)
		tinsert(self._traced, frame);
	end,
	["UnregisterTracer"] = function(self, frame)
		for index, f in ipairs(self._traced) do
			if frame == f then
				tremove(self._traced, index);
				break;
			end
		end
		self._stateMap[frame] = nil;
	end,
};

function MessCursorTracer_OnLoad(self)
	inherits(self, MessCursorTracerMeta);
	self._traced = {};
	self._stateMap = {};
	self.tracer:Play();
end

function MessCursorTracer_OnTime(self, state)
	local parent = self:GetParent();
	for index, frame in ipairs(parent._traced) do
		if frame:IsVisible() then
			local newState = frame:IsMouseOver() and Frame_OnEnter or Frame_OnLeave;
			if parent._stateMap[frame] ~= newState then
				parent._stateMap[frame] = newState;
				pcall(frame[newState], frame);
			end
		else
			parent._stateMap[frame] = nil;
		end
	end
end
