local mathMax = math.max;
local mathMin = math.min;
local mathFloor = math.floor;
local mathCeil = math.ceil;

local FADEOUTTIME = 3;
local MAXIUMWIDTH = 300;
local MAXIUMHEIGHT = 500;
local DEFAULTWIDTH = 100;
local ITEMPADDING = 5;

local SCROLLITEMUPLIMIT = 6;

local COLOR_FOREGROUND = {0.196, 0.196, 0.196};
local COLOR_BACKGROUND = {0.9, 0.9, 0.9};
local COLOR_DISABLED = {0.4, 0.4, 0.4};

local ICONTITLE = "INTERFACE\\Vehicles\\UI-VEHICLES-RAID-ICON";
local ICONCHECKED = "INTERFACE\\Challenges\\challenges-copper";
local ICONMORE = "INTERFACE\\ChatFrame\\ChatFrameExpandArrow";

local REF_CURSOR = "cursor";
local REF_REGION = "region";

local function setVisible(region, visible)
	if visible then
		region:Show();
	else
		region:Hide();
	end
end

local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

local function showAlongside(self, relative, offsetX, offsetY)
	offsetX = math.abs(offsetX or 0);
	offsetY = math.abs(offsetY or 0);
	local sw, sh = GetScreenWidth(), GetScreenHeight();
	
	self:ClearAllPoints();
	if not relative then
		local cursorX, cursorY = GetCursorPosition();
		self:SetPoint(
			string.format("%s%s", cursorY >= sh / 2 and "TOP" or "BOTTOM", cursorX >= sw / 2 and "RIGHT" or "LEFT"),
			UIParent, "BOTTOMLEFT", cursorX - UIParent:GetLeft(), cursorY - UIParent:GetBottom(),
			cursorX >= sw / 2 and -offsetX or offsetX,
			cursorY >= sh / 2 and -offsetY or offsetY
		);
	else
		local x, y, w, h = relative:GetLeft(), relative:GetBottom(), relative:GetSize();
		local centerX, centerY = x + w / 2, y + h / 2;
		self:SetPoint(
			string.format("%s%s", centerY >= sh / 2 and "TOP" or "BOTTOM", centerX >= sw / 2 and "RIGHT" or "LEFT"),
			relative,
			string.format("%s%s", centerY >= sh / 2 and "TOP" or "BOTTOM", centerX >= sw / 2 and "LEFT" or "RIGHT"),
			centerX >= sw / 2 and -offsetX or offsetX,
			centerY >= sh / 2 and -offsetY or offsetY
		);
	end
	self:Show();
	self:Raise();
end

----------------------------------------------------------------	temples
--[[
--	data structure to feed for item, it depends on itemTemplate Visualize implementation.
--	state is simliar structure, but one time only.
dataTemple = {
	--	basic
	id,			--	unique id, used to identify state
	text,		--	text of menu item
	func,		--	callback for item clicked
	type,		--	item type, e.g. normal, title, checker, indicator
	--	optional
	prefixIcon,	--	replace icon for prefix
	suffixIcon,	--	replace icon for suffix
	isDisabled,	--	if item in disabled state
	isChecked,	--	if item in checked state, only works under checker type
	submenu,	--	for submenu data table, only works under indicator type
}
]]
----------------------------------------------------------------	MenuMouseoverTimer
function MenuMouseoverTimer_OnUpdate(self, elpsed)
	local parent = self:GetParent();
	if parent:IsMouseOver() then
		parent.countdownBar:HaltCountdown();
	else
		parent.countdownBar:BeginCountdown(self._fadeoutTime);
	end
end

----------------------------------------------------------------	MessMenuItemTemplate
local MessMenuItemMeta = {
	["_menuParent"] = nil,
	["SetMenuParent"] = function(self, parent)
		self._menuParent = parent;
	end,
	["GetMenuParent"] = function(self)
		return self._menuParent or self:GetParent();
	end,
	["onClickFunc"] = nil,
}

function MessMenuItem_OnLoad(self)
	inherits(self, MessMenuItemMeta);
end

function MessMenuItem_OnEnter(self, ...)
	local parent = self:GetMenuParent();
	self.text:SetTextColor(unpack(parent._backgroundColor));
	if self:IsEnabled() then
		self.bg:SetVertexColor(unpack(parent._foregroundColor));
	else
		self.bg:SetVertexColor(unpack(parent._diabledColor));
	end
end

function MessMenuItem_OnLeave(self, ...)
	local parent = self:GetMenuParent();
	if self:IsEnabled() then
		self.text:SetTextColor(unpack(parent._foregroundColor));
	else
		self.text:SetTextColor(unpack(parent._diabledColor));
	end
	self.bg:SetVertexColor(unpack(parent._backgroundColor));
end

function MessMenuItem_OnClick(self, ...)
	pcall(self.onClickFunc, self, ...);
end

function MessMenuItem_PostClick(self, ...)
	local parent = self:GetMenuParent();
	pcall(parent.itemSelectedFunc, parent, self);
end

----------------------------------------------------------------	MessMenuTemplate
local MessMenuMeta = {
	["_dataCache"] = nil,
	["_itemTemplate"] = "MessMenuItemTemplate",
	["_autoSizing"] = true,
	["_preferedWidth"] = DEFAULTWIDTH,
	["_maxiumWidth"] = MAXIUMWIDTH,
	["_persisted"] = false,
	["_fadeoutTime"] = FADEOUTTIME,
	["_foregroundColor"] = COLOR_FOREGROUND,
	["_backgroundColor"] = COLOR_BACKGROUND,
	["_diabledColor"] = COLOR_DISABLED,
	
	["iconTitle"] = ICONTITLE,
	["iconChecked"] = ICONCHECKED,
	["iconMore"] = ICONMORE,
	--	methods
	["_getItem"] = function(self)
		local item = tremove(self._itemPool);
		if not item then
			item = CreateFrame("Button", string.format("$parentItem_%d", self._poolCount), self, self._itemTemplate);
			self._poolCount = self._poolCount + 1;
		end
		item:SetMenuParent(self);
		return item;
	end,
	["_returnItem"] = function(self, item)
		item:ClearAllPoints();
		item:Hide();
		tinsert(self._itemPool, item);
	end,
	["SetItemTemplate"] = function(self, templateName)	-- once change this attribute, menu need to be feeded again.
		self._itemTemplate = value;
		self:Close();
		for _, item in ipairs(self._itemList) do
			self:_returnItem(item);
		end
		wipe(self._itemPool);
	end,
	["SetAutosizing"] = function(self, flag)
		self._autoSizing = flag;
	end,
	["SetPreferedWidth"] = function(self, width)
		self._preferedWidth = width;
	end,
	["SetMaxiumWidth"] = function(self, width)
		self._maxiumWidth = width;
	end,
	["SetPersisted"] = function(self, flag)
		self._persisted = flag;
	end,
	["SetForegroundColor"] = function(self, color)
		color = color or COLOR_FOREGROUND;
		self.LBorder:SetVertexColor(unpack(color));
		self.RBorder:SetVertexColor(unpack(color));
		self.BBorder:SetVertexColor(unpack(color));
		self.TBorder:SetVertexColor(unpack(color));
		self.countdownBar:SetIndicatorColor(color);
		self._foregroundColor = color;
		self:Update();
	end,
	["SetBackgroundColor"] = function(self, color)
		self._backgroundColor = color or COLOR_BACKGROUND;
		self:Update();
	end,
	["SetDisabledColor"] = function(self, color)
		self._diabledColor = color or COLOR_DISABLED;
		self:Update();
	end,
	["SetFadeoutTime"] = function(self, duration)
		self._fadeoutTime = duration;
		pcall(self:GetScript(self:IsMouseOver() and "OnEnter" or "OnLeave"), self);
	end,
	["FeedDataTable"] = function(self, dataTab)	--	dataTab	>	index : data
		assert(type(dataTab) == "table");

		for _, item in ipairs(self._itemList) do
			self:_returnItem(item);
		end
		
		local wMax, hMax, ptr = 0, 0;
		for index, data in ipairs(dataTab) do
			local item = self:_getItem();
			item.index = data.index;
			local w, h = self:Visualize(item, data);
			item:SetPoint("LEFT");
			item:SetPoint("RIGHT");
			if ptr then
				item:SetPoint("TOP", ptr, "BOTTOM");
			else
				item:SetPoint("TOP");
			end
			item:Show();

			tinsert(self._itemList, item);
			wMax = mathMax(wMax, w);
			hMax = hMax + h;
			ptr = item;
		end
		self._dataCache = dataTab;

		self:SetHeight(hMax);
		if self._autoSizing then
			self:SetWidth(mathMin(self._maxiumWidth, wMax));
		else
			self:SetWidth(self._preferedWidth);
		end
	end,
	["GetData"] = function(self, index)
		return self._dataCache[index];
	end,
	["Open"] = function(self, reference, ...)
		showAlongside(self, reference);
	end,
	["Close"] = function(self)
		self:Hide();
	end,
	["Update"] = function(self, stateTab)				--	stateTab : id : state
		if stateTab then
			for _, data in ipairs(self._dataCache) do
				if stateTab[data.id] then
					for key, value in pairs(stateTab[data.id]) do
						data[key] = value;
					end
				end
			end
		else
			for index, item in ipairs(self._itemList) do
				self:Visualize(item, self._dataCache[index]);
				pcall(item:IsMouseOver() and item:GetScript("OnEnter") or item:GetScript("OnLeave"), item);
			end
		end
	end,
	["Visualize"] = function(self, item, data)
		data = data or self:GetData(item.index);
		assert(data);
		--	basic
		item.text:SetText(data.text);
		item.onClickFunc = data.func;
		--	pre/suffix
		if data.prefixIcon then
			item.prefixTexture:SetTexture(data.prefixIcon);
		elseif data.type == "TITLE" then
			item.prefixTexture:SetTexture(self.iconTitle);
		elseif data.type == "CHECKER" then
			item.prefixTexture:SetTexture(data.isChecked and self.iconChecked);
		else
			item.prefixTexture:SetTexture(nil);
		end
		if data.suffixIcon then
			item.suffixTexture:SetTexture(data.suffixIcon);
		elseif data.type == "INDICATOR" then
			item.suffixTexture:SetTexture(self.iconMore);
		else
			item.suffixTexture:SetTexture(nil);
		end
		--	disable
		item:SetEnabled(not data.isDisabled);
		--	sizing
		return item.prefixTexture:GetWidth() + item.suffixTexture:GetWidth() + item.text:GetStringWidth(), item:GetHeight();
	end,
	["itemSelectedFunc"] = function(self, item)			--	for MessMenuItem
		if self._persisted then
			self:Update();
		else
			self:Close();
		end
	end,
	["Counter_OnFinish"] = function(self, trigger)
		self.countdownBar:Hide();
		self:GetParent().fadeout:Play();
	end,
};

function MessMenu_OnLoad(self)
	inherits(self, MessMenuMeta);

	self._itemPool = {};
	self._poolCount = 0;
	self._itemList = {};
	
	self:SetForegroundColor(COLOR_FOREGROUND);
	self:SetBackgroundColor(COLOR_BACKGROUND);
	self:SetDisabledColor(COLOR_DISABLED);
	self:SetFadeoutTime(FADEOUTTIME);
	
	self.countdownBar:RegisterCallback(self, "Counter_OnFinish");
end

function MessMenu_OnShow(self)
	self.fadein:Play();
end

function MessMenu_OnEnter(self, motion)
	if self._fadeoutTime and self._fadeoutTime > 0 then
		self.mouseoverTimer:Play();
	end
end

function MessMenu_OnLeave(self, motion)
	self.mouseoverTimer:Stop();
end

----------------------------------------------------------------	MessFauxScrollMenuTemplate
local MessFauxScrollMenuMeta = {
	["_dataList"] = nil,
	["_itemTemplate"] = "MessMenuItemTemplate",
	["_autoSizing"] = true,
	["_preferedWidth"] = DEFAULTWIDTH,
	["_maxiumWidth"] = MAXIUMWIDTH,
	["_persisted"] = false,
	["_fadeoutTime"] = FADEOUTTIME,
	["_foregroundColor"] = COLOR_FOREGROUND,
	["_backgroundColor"] = COLOR_BACKGROUND,
	["_diabledColor"] = COLOR_DISABLED,
	
	["iconTitle"] = ICONTITLE,
	["iconChecked"] = ICONCHECKED,
	["iconMore"] = ICONMORE,
	--	functions
	["SetItemTemplate"] = function(self, templateName)	-- once change this attribute, menu need to be feeded again.
		self._itemTemplate = value;
		self:Close();
		self:Init();
	end,
	["SetAutosizing"] = function(self, flag)
		self._autoSizing = flag;
	end,
	["SetPreferedWidth"] = function(self, width)
		self._preferedWidth = width;
	end,
	["SetMaxiumWidth"] = function(self, width)
		self._maxiumWidth = width;
	end,
	["SetPersisted"] = function(self, flag)
		self._persisted = flag;
	end,
	["SetForegroundColor"] = function(self, color)
		color = color or COLOR_FOREGROUND;
		self.LBorder:SetVertexColor(unpack(color));
		self.RBorder:SetVertexColor(unpack(color));
		self.BBorder:SetVertexColor(unpack(color));
		self.TBorder:SetVertexColor(unpack(color));
		self.countdownBar:SetIndicatorColor(color);
		self._foregroundColor = color;
		self:Update();
	end,
	["SetBackgroundColor"] = function(self, color)
		self._backgroundColor = color or COLOR_BACKGROUND;
		self:Update();
	end,
	["SetDisabledColor"] = function(self, color)
		self._diabledColor = color or COLOR_DISABLED;
		self:Update();
	end,
	["SetFadeoutTime"] = function(self, duration)
		self._fadeoutTime = duration;
		pcall(self:GetScript(self:IsMouseOver() and "OnEnter" or "OnLeave"), self);
	end,
	["Init"] = function(self)
		self.scrollpane:Init(self._itemTemplate, "Button");
		for _, item in ipairs(self.scrollpane.items) do
			item:SetMenuParent(self);
		end
	end,
	["FeedDataTable"] = function(self, dataTab)
		assert(type(dataTab) == "table");

		self._dataList = dataTab;
		self._dataCount = #dataTab;

		self:SetHeight(self.scrollpane.unitHeight * mathMin(self._dataCount, SCROLLITEMUPLIMIT));
		if self._autoSizing then
			local wMax = 0;
			local sample = self.scrollpane.items[1];
			for index, data in ipairs(dataTab) do
				local w, h = self:Visualize(sample, data);
				wMax = mathMax(wMax, w);
			end
			self:SetWidth(mathMin(self._maxiumWidth, wMax));
		else
			self:SetWidth(self._preferedWidth or DEFAULTWIDTH);
		end

		self:Update();
	end,
	["GetData"] = function(self, index)
		return self._dataList[index];
	end,
	["Open"] = function(self, reference, ...)
		showAlongside(self, reference);
	end,
	["Close"] = function(self)
		self:Hide();
	end,
	["Update"] = function(self, stateTab)
		if stateTab then
			for _, data in ipairs(self._dataList) do
				if stateTab[data.id] then
					for key, value in pairs(stateTab[data.id]) do
						data[key] = value;
					end
				end
			end
		end
		self.scrollpane:Update();
	end,
	["Visualize"] = function(self, item, data)
		data = data or self:GetData(item.index);
		assert(data);
		--	basic
		item.text:SetText(data.text);
		item.onClickFunc = data.func;
		--	pre/suffix
		if data.prefixIcon then
			item.prefixTexture:SetTexture(data.prefixIcon);
		elseif data.type == "TITLE" then
			item.prefixTexture:SetTexture(self.iconTitle);
		elseif data.type == "CHECKER" then
			item.prefixTexture:SetTexture(data.isChecked and self.iconChecked);
		else
			item.prefixTexture:SetTexture(nil);
		end
		if data.suffixIcon then
			item.suffixTexture:SetTexture(data.suffixIcon);
		elseif data.type == "INDICATOR" then
			item.suffixTexture:SetTexture(self.iconMore);
		else
			item.suffixTexture:SetTexture(nil);
		end
		--	disable
		item:SetEnabled(not data.isDisabled);
		--	sizing
		return item.prefixTexture:GetWidth() + item.suffixTexture:GetWidth() + item.text:GetStringWidth(), item:GetHeight();
	end,
	["itemSelectedFunc"] = function(self, item)		--	for MessMenuItem
		if self._persisted then
			self:Update();
		else
			self:Close();
		end
	end,
	["Counter_OnFinish"] = function(self, trigger)
		self:GetParent().fadeout:Play();
	end,
};

local function GetFauxOffsetItem(scrollpane, offsetPixel)
	local parent = scrollpane:GetParent();
	for index, data in ipairs(parent._dataList) do
		offsetPixel = offsetPixel - scrollpane.unitHeight;
		if offsetPixel <= 0 then
			return index, offsetPixel + scrollpane.unitHeight;
		end
	end
end

local function UpdateFauxScrollpane(scrollpane)
	if not scrollpane:IsVisible() then return; end

	local parent = scrollpane:GetParent();
	if parent._dataCount == 0 then
		scrollpane.bgText:Show();
	else
		scrollpane.bgText:Hide();

		local offset = scrollpane:GetTopItemIndex();
		for i = 1, #scrollpane.items do
			local item = scrollpane.items[i];
			local index = offset + i - 1;
			item.index = index;
			if index <= parent._dataCount then
				parent:Visualize(item, parent._dataList[index]);
				item:Show();
			else
				item:Hide();
			end
		end
	end
end

function MessScrollMenu_OnLoad(self)
	inherits(self, MessFauxScrollMenuMeta);

	self._dataList = {};
	self._dataCount = 0;
	
	local scrollpane = self.scrollpane;
	scrollpane.GetOffsetItem = GetFauxOffsetItem;
	scrollpane.Update = UpdateFauxScrollpane;
	
	local bar = self.scrollpane.bar;
	bar:SetWidth(2);
	bar:SetBorderVisible(false);
	bar:SetFading(false);
	bar:SetControlable(false);
	bar:SetIndicatorColor(COLOR_FOREGROUND);
	
	self:SetForegroundColor(COLOR_FOREGROUND);
	self:SetBackgroundColor(COLOR_BACKGROUND);
	self:SetDisabledColor(COLOR_DISABLED);
	self:SetFadeoutTime(FADEOUTTIME);
	self.countdownBar:RegisterCallback(self, "Counter_OnFinish");
end

function MessScrollMenu_OnShow(self)
	self.fadein:Play();
	self.scrollpane.VerticalSmoothScrollTo(0);	-- rewind
end

function MessScrollMenu_OnSizeChanged(self, w, h)
	self:Init();
end
