local mathMax = math.max;
local mathMin = math.min;
local mathFloor = math.floor;
local mathCeil = math.ceil;
local PI = math.pi;

local THRESHOLD = 0.01;

local function setVisible(region, visible)
	if visible then
		region:Show();
	else
		region:Hide();
	end
end

local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

local function pageButtonOnEnter(self, motion)
	local parent = self:GetParent();
	parent.subtitle:SetText(self.text);
end

local function pageButtonOnLeave(self, motion)
	local parent = self:GetParent();
	parent.subtitle:SetText(parent.subtitle.text);
end

----------------------------------------------------------------	MessGlassPaneTemplate
function MessGlassPane_OnShow(self)
	self.fadeIn:Play();
end

----------------------------------------------------------------	MessPanelTemplate
local MessPanelMeta = {
	["noFading"] = false,
	["_fadeoutFinished"] = function(self, ...)
		self:GetParent():Hide();
	end,
	["Active"] = function(self, skipAnim)
		if self:IsShown() then return; end
		self:StopAnimating();
		self:Show();
		if not self.noFading and not skipAnim then
			self.fadein:Play();
		end
		if self.onActiveFunc then
			self:onActiveFunc();
		end
	end,
	["Deactive"] = function(self, skipAnim)
		if not self:IsShown() then return; end
		self:StopAnimating();
		if not self.noFading and not skipAnim then
			self.fadeout:Play();
		else
			self:Hide();
		end
		if self.onDeactiveFunc then
			self:onDeactiveFunc();
		end
	end,
	["SetTitle"] = function(self, title, subtitle)
		self.title:SetText(title);
		self.subtitle.text = subtitle;
		self.subtitle:SetText(subtitle);
	end,
	["SetPrevious"] = function(self, icon, text, func)
		self.previous.text = text;
		self.previous:SetIcon(icon);
		setVisible(self.previous, icon);
		self.previous:SetScript("OnClick", func);
		self.previous:HookScript("OnEnter", pageButtonOnEnter);
		self.previous:HookScript("OnLeave", pageButtonOnLeave);
	end,
	["SetNext"] = function(self, icon, text, func)
		self.next.text = text;
		self.next:SetIcon(icon);
		setVisible(self.next, icon);
		self.next:SetScript("OnClick", func);
		self.next:HookScript("OnEnter", pageButtonOnEnter);
		self.next:HookScript("OnLeave", pageButtonOnLeave);
	end,
};

function MessPanel_OnLoad(self)
	inherits(self, MessPanelMeta);
	
	self.fadeout:SetScript("OnFinished", self._fadeoutFinished);
end

----------------------------------------------------------------	MessModelDisplayTemplate
local MessModelDisplayMeta = {
	["autoRotate"] = true,
	["SetModelAnime"] = function(self, index)
		if not index or index == 0 then
			self:SetSequence(0);
			self._anime = nil;
		else
			self._anime = index;
		end
	end,
	["GetModelAnime"] = function(self)
		return self._anime;
	end,
};

function MessModelDisplay_OnLoad(self)
	inherits(self, MessModelDisplayMeta);
	
	self:RegisterForDrag("LeftButton", "RightButton");
	self:SetUnit("player");
	--self:SetPosition(-0.5, 0, 0);
end

function MessModelDisplay_OnShow(self)
	self._seqtime = 0;
end

function  MessModelDisplay_OnAnimFinished(self)
	self._seqtime = 0;
end

function MessModelDisplay_OnMouseWheel(self, delta)
	local px, py, pz = self:GetPosition();
	self:SetPosition(px + delta * -0.005, py, pz);
end

function  MessModelDisplay_OnDragStart(self)
	self._cursorX, self._cursorY = GetCursorPosition();
	self._dragging = button;
end

function  MessModelDisplay_OnDragStop(self)
	self._dragging = false;
end

function  MessModelDisplay_OnUpdate(self, elapsed)
	self._elapsed = self._elapsed + elapsed;
	if THRESHOLD > self._elapsed then return end
	elapsed = mathFloor(self._elapsed / THRESHOLD) * THRESHOLD;
	self._elapsed = self._elapsed - elapsed;
	
	-- animation
	if self._anime then
		self._seqtime = self._seqtime + elapsed * 1000; 
		self:SetSequenceTime(self._anime, self._seqtime);
	end
	
	if self._dragging then
		local cursorX, cursorY = GetCursorPosition();
		local x, y = self._cursorX - cursorX, self._cursorY - cursorY;
		if self._dragging == "LeftButton" then
			self:SetFacing(mathMax(mathMin(self:GetFacing() + x * -0.005, PI), -PI));
			local px, py, pz = self:GetPosition();
			self:SetPosition(px + y * -0.005, py, pz);
		elseif self._dragging == "RightButton" then	
			local px, py, pz = self:GetPosition();
			self:SetPosition(px, py + x * -0.005, pz + y * -0.005);
		end
		self._cursorX, self._cursorY = GetCursorPosition();
	elseif self.autoRotate then -- rotate
		self:SetFacing( self:GetFacing() + elapsed * PI / 4 );
	end
end

----------------------------------------------------------------	MessAlertTemplate
local MessAlertMeta = {
	["Alert"] = function(self, icon, title, subtitle)
		self.icon:SetTexture(icon);
		self.title:SetText(title);
		self.subtitle:SetText(subtitle);
		self:Show();
	end,
};

function MessAlert_OnLoad(self)
	inherits(self, MessAlertMeta);
	
	self:RegisterForDrag("LeftButton", "RightButton");
end

function MessAlert_OnShow(self)
	self.glow:Show();
	self.shine:Show();
	self.fadeIn:Play();
	
	if self.onActiveFunc then
		self:onActiveFunc();
	end
end

function MessAlert_OnHide(self)
	if self.onDeactiveFunc then
		self:onDeactiveFunc();
	end
end

function MessAlert_OnDragStart(self)
	self:StartMoving();
end

function MessAlert_OnDragStop(self)
	self:StopMovingOrSizing();
end

function MessAlert_OnEnter(self, motion)
	if motion and not self.noHoldback then
		self.fadeOut:Stop();
	end
end

function MessAlert_OnLeave(self, motion)
	self.fadeOut:Play();
end
