local THRESHOLD = 0.01;
local SCROLLTIME = 0.3;
local TIMETOCOUNTDOWN = 3;
local SCROLLSTEP = 20;

local ORI_VERTICAL = "VERTICAL";
local ORI_HORIZONTAL = "HORIZONTAL";
local Counter_OnBegin = "Counter_OnBegin";
local Counter_OnFinish = "Counter_OnFinish";

local COLOR_BORDER = {0.5, 0.5, 0.5, 1};
local COLOR_INDICATOR = {1, 0.5, 0, 1};
local COLOR_INDICATORACTIVED = {1, 0.75, 0, 1};

local mathMax = math.max;
local mathMin = math.min;
local mathFloor = math.floor;
local mathCeil = math.ceil;

local function mathRound(num)
	return mathFloor(num + .5);
end

local function mathBound(value, min, max)
	return mathMax(min, mathMin(max, value));
end

local function setVisible(region, visible)
	if visible then
		region:Show();
	else
		region:Hide();
	end
end

local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

----------------------------------------------------------------	MessScrollBarTemplate
local MessScrollBarMeta = {
	--	attributes
	["_scrollParent"] = nil,
	["_controlable"] = true,
	["_viewportWidth"] = 1,
	["_viewportHeight"] = 1,
	["_oldOrientation"] = ORI_VERTICAL,
	["_preferSmooth"] = true,
	["_doFading"] = true,
	["borderColor"] = COLOR_BORDER,
	["indicatorColor"] = COLOR_INDICATOR,
	["indicatorActiveColor"] = COLOR_INDICATORACTIVED,
	["activeAlpha"] = 1.0,
	["idleAlpha"] = 0.0,
	--	functions
	["_raiseAlpha"] = function(self)
		if not self:IsFading() then
			--	nothing
		elseif self.fadein:IsPlaying() then
			--	fading in, let it finish
		elseif self.fadeout:IsPlaying() then
			--	fading out, rewind it and play
			self.fadeout:Stop();
			if not self._indicatorInteracting and not self:IsMouseOver() then
				self.fadeout:Play();
			end
		elseif self:GetAlpha() == self.activeAlpha then
			--	ativing, do nothing
		else
			--	idling, begin fade in
			self.fadein:Play();
		end
	end,
	["_lowerAlpha"] = function(self)
		if not self:IsFading() then
			--	nothing
		elseif self._indicatorInteracting or self:IsMouseOver() then
			--	nothing
		elseif self.fadein:IsPlaying() then
			--	fading in, let it finish & trigger fadeout
		elseif self.fadeout:IsPlaying() then
			--	fading out, let it finish
		elseif self:GetAlpha() == self.activeAlpha then
			self.fadeout:Play();
		else
			--	idling, do nothing
		end
	end,
	["Update"] = function(self)
		--	if orientation chaned
		if self:GetOrientation() ~= self._oldOrientation then
			self.indicator:ClearAllPoints();
			if self:GetOrientation() == ORI_VERTICAL then
				self.indicator:SetPoint("LEFT", 1, 0);
				self.indicator:SetPoint("RIGHT", -1, 0);
			else
				self.indicator:SetPoint("TOP", -1, 0);
				self.indicator:SetPoint("BOTTOM", 1, 0);
			end
			self._oldOrientation = self:GetOrientation();
		end
		--	update indicator
		local min, max = self:GetMinMaxValues();
		local range = max - min;
		local value = self:GetValue();
		local vw, vh = self:GetViewportSize();
		if self:GetOrientation() == ORI_VERTICAL then
			local space = self:GetHeight() - 2;
			local h, y;
			if mathRound(range) == 0 then
				h = space;
				y = 0;
			else
				h = mathMax(1, space * vh / (range + vh));
				y = (space - h) * value / range;
			end
			self.indicator:SetHeight(h);
			self.indicator:SetPoint("TOP", self, "TOP", 0, -1 - y);
		else
			local space = self:GetWidth() - 2;
			local w, x;
			if mathRound(range) == 0 then
				w = space;
				x = 0;
			else
				w = mathMax(1, space * vw / (range + vw));
				x = (space - w) * value / range;
			end
			self.indicator:SetWidth(w);
			self.indicator:SetPoint("LEFT", self, "LEFT", 1 + x, 0);
		end
		self:_raiseAlpha();
	end,
	["GetViewportSize"] = function(self)
		return self._viewportWidth, self._viewportHeight;
	end,
	["SetViewportSize"] = function(self, w, h)
		self._viewportWidth = w;
		self._viewportHeight = h;
		self:Update();
	end,
	["SetScrollParent"] = function(self, parent)
		self._scrollParent = parent;
	end,
	["GetScrollParent"] = function(self)
		return self._scrollParent or self:GetParent();
	end,
	["SetBorderColor"] = function(self, color)
		color = color or COLOR_BORDER;
		self.LBorder:SetVertexColor(unpack(color));
		self.RBorder:SetVertexColor(unpack(color));
		self.BBorder:SetVertexColor(unpack(color));
		self.TBorder:SetVertexColor(unpack(color));
		self.borderColor = color;
	end,
	["SetBorderVisible"] = function(self, visible)
		if visible then
			self:EnableDrawLayer("BORDER");
		else
			self:DisableDrawLayer("BORDER");
		end
	end,
	["SetIndicatorColor"] = function(self, color, activeColor)
		color = color or COLOR_INDICATOR;
		activeColor = activeColor or COLOR_INDICATORACTIVED;
		self.indicator:SetVertexColor(unpack(color));
		self.indicatorColor = color;
		self.indicatorActiveColor = activeColor;
	end,
	["SetFading"] = function(self, fading)
		self._doFading = fading;
	end,
	["IsFading"] = function(self)
		return self._doFading;
	end,
	["SetFadeAlpha"] = function(self, active, idle)
		self.activeAlpha = mathBound(active, 0, 1.0);
		self.idleAlpha = mathBound(idle, 0, 1.0);
		self.fadein.anim:SetChange(self.activeAlpha - self.idleAlpha);
		self.fadeout.anim:SetChange(self.idleAlpha - self.activeAlpha);
		if self.activeAlpha == self.idleAlpha then
			self:SetFading(false);
		end
		self:SetAlpha(self.idleAlpha);
	end,
	["GetFadeAlpha"] = function(self)
		return self.activeAlpha, self.idleAlpha;
	end,
	["SetFadeDuration"] = function(self, fadein, fadeout, timeTofade)
		self.fadein.anim:SetDuration(fadein);
		self.fadeout.anim:SetDuration(fadeout);
		self.fadeout.anim:SetStartDelay(timeTofade);
		if timeTofade == 0 then
			self:SetFading(false);
		end
	end,
	["GetFadeDuration"] = function(self)
		return self.fadein.anim:GetDuration(), self.fadeout.anim:GetDuration(), self.fadeout.anim:GetStartDelay();
	end,
	["SetPreferSmooth"] = function(self, enabled)
		self._preferSmooth = enabled;
	end,
	["IsPreferSmooth"] = function(self)
		return self._preferSmooth;
	end,
	["SetControlable"] = function(self, enabled)
		self:EnableMouse(enabled);
		self:EnableMouseWheel(enabled);
		self:RegisterForDrag(enabled and "LeftButton");
		self._controlable = enabled;
	end,
	["IsControlable"] = function(self)
		return self._controlable;
	end,
};

function MessScrollBar_OnLoad(self)
	inherits(self, MessScrollBarMeta);
	self:SetValueStep(0.01);
	self:RegisterForDrag("LeftButton");
	
	self:SetBorderColor(COLOR_BORDER);
	self:SetIndicatorColor(COLOR_INDICATOR, COLOR_INDICATORACTIVED);
	self:SetAlpha(self.idleAlpha);
			
	hooksecurefunc(self, "SetOrientation", function(self, orient)
		self:Update();
	end);
end

function MessScrollBar_OnShow(self)
	self:Update();
end

function MessScrollBar_OnEnter(self, motion)
	self:_raiseAlpha();
end

function MessScrollBar_OnLeave(self, motion)
	self:_lowerAlpha();
end

function MessScrollBar_Update(self)
	self:Update();
end

function MessScrollBar_OnMouseDown(self, button)
	-- if cursor is inside indicator
	if self.indicator:IsMouseOver(0, 0, 0, 0) then
		-- prepare for OnDragStart
		self._indicatorInteracting = true;
		self:_raiseAlpha();
		local x, y = GetCursorPosition();
		self.dragOffsetX = x - self.indicator:GetLeft();
		self.dragOffsetY = self.indicator:GetTop() - y;
		self.indicator:SetVertexColor(unpack(self.indicatorActiveColor));
	end
end

function MessScrollBar_OnMouseUp(self, button)
	--	OnDragStop will not trigger OnMouseUp
	if self._indicatorInteracting then	
		--	case : click on indicator without triggering OnDragStart
		self._indicatorInteracting = nil;
		self.indicator:SetVertexColor(unpack(self.indicatorColor));
	else
		--	scroll to click position
		local parent = self:GetScrollParent();
		local min, max = self:GetMinMaxValues();
		local range = max - min;
		local orientation = self:GetOrientation();
		local x, y = GetCursorPosition();
		-- scroll indicator center to cursor position
		if orientation == ORI_VERTICAL then
			local space = self:GetHeight() - 2;
			local h = self.indicator:GetHeight();
			local value = (self:GetTop() - y - h / 2) / (space - h) * range;
			if self:IsPreferSmooth() and parent.VerticalSmoothScrollTo then
				parent:VerticalSmoothScrollTo(value);
			else
				parent:SetScrollValue(value, ORI_VERTICAL);
			end
		else
			local space = self:GetWidth() - 2;
			local w = self.indicator:GetWidth();
			local value = (x - self:GetLeft() - w / 2) / (space - w) * range;
			if self:IsPreferSmooth() and parent.HorizontalSmoothScrollTo then
				parent:HorizontalSmoothScrollTo(value);
			else
				parent:SetScrollValue(value, ORI_HORIZONTAL);
			end
		end
	end
	self:_lowerAlpha();
end

function MessScrollBar_OnDragStart(self, button)
	if self._indicatorInteracting then
		self.draggingTimer:Play();
	end
end

function MessScrollBar_OnDragStop(self)
	if self._indicatorInteracting then
		self._indicatorInteracting = nil;
		self.draggingTimer:Stop();
		self.indicator:SetVertexColor(unpack(self.indicatorColor));
		self:_lowerAlpha();
	end
end

function MessScrollBar_OnMouseWheel(self, delta)
	local parent = self:GetScrollParent();
	if self:GetOrientation() == ORI_VERTICAL then
		if self:IsPreferSmooth() and parent.VerticalSmoothDeltaScroll then
			parent:VerticalSmoothDeltaScroll(delta);
		else
			local step = parent.GetScrollStep and parent:GetScrollStep() or parent.scrollStep;
			local value = self:GetValue() - delta * step;
			parent:SetScrollValue(mathBound(value, self:GetMinMaxValues()), ORI_VERTICAL);
		end
	else
		if self:IsPreferSmooth() and parent.HorizontalSmoothDeltaScroll then
			parent:HorizontalSmoothDeltaScroll(delta);
		else
			local step = parent.GetScrollStep and parent:GetScrollStep() or parent.scrollStep;
			local value = self:GetValue() - delta * step;
			parent:SetScrollValue(mathBound(value, self:GetMinMaxValues()), ORI_HORIZONTAL);
		end
	end
	--	self:_raiseAlpha will be called by MessScrollBar_Update
end

function MessScrollBar_FadeinFinished(self)
	bar = self:GetParent();
	bar:SetAlpha(bar.activeAlpha);
	bar:_lowerAlpha();
end

function MessScrollBar_FadeoutFinished(self)
	bar = self:GetParent();
	bar:SetAlpha(bar.idleAlpha);
end

function MessScrollBar_DraggingUpdate(self, elapsed)
	local bar = self:GetParent();
	-- assume minmax, viewportsize won't change while dragging. if it happens, things go wried :3
	local parent = bar:GetScrollParent();
	local min, max = bar:GetMinMaxValues();
	local range = max - min;
	local orientation = bar:GetOrientation();
	local scale, x, y = bar:GetEffectiveScale(), GetCursorPosition();
	--	dragging handler, no smooth on dragging guestures
	if orientation == ORI_VERTICAL then
		local h = bar.indicator:GetHeight();
		local value = (bar:GetTop() - 1 - (y + bar.dragOffsetY) / scale) / (bar:GetHeight() - 2 - h) * range;
		parent:SetScrollValue(mathBound(value, min, max), ORI_VERTICAL);
	else
		local w = bar.indicator:GetWidth();
		local value = ((x - bar.dragOffsetX) / scale - bar:GetLeft() - 1) / (bar:GetWidth() - 2 - w) * range;
		parent:SetScrollValue(mathBound(value, min, max), ORI_HORIZONTAL);
	end
end

----------------------------------------------------------------	MessCountdownBarTemplate
local MessCountdownBarMeta = {
	--	attributes
	["_scrollParent"] = nil,
	["_orientation"] = ORI_HORIZONTAL,
	--	methods
	["SetScrollParent"] = function(self, parent)
		self._scrollParent = parent;
	end,
	["GetScrollParent"] = function(self)
		return self._scrollParent or self:GetParent();
	end,
	["GetOrientation"] = function(self)
		return self._orientation;
	end,
	["SetOrientation"] = function(self, orient)
		assert(orient ~= ORI_VERTICAL and orient ~= ORI_HORIZONTAL, "MessCountdownBar : unknown orientation" .. orient);
		self._orientation = orient;
	end,
	["SetIndicatorColor"] = function(self, color)
		color = color or COLOR_INDICATOR;
		self.indicator:SetVertexColor(unpack(color));
	end,
	["BeginCountdown"] = function(self, duration)
		if self.countdownTimer:IsPlaying() then	return;	end
		self:Show();
		self.countdownTimer.anim:SetDuration(duration or TIMETOCOUNTDOWN);
		self.countdownTimer:Play();
	end,
	["HaltCountdown"] = function(self)
		self.countdownTimer:Stop();
	end,
	["IsCountdown"] = function(self)
		return self.countdownTimer:IsPlaying();
	end,
};

function MessCountdownBar_OnLoad(self)
	MessEventDispatcher_OnLoad(self);
	inherits(self, MessCountdownBarMeta);
end

function MessCountdownBar_OnHide(self)
	self.countdownTimer:Stop();
end

function MessCountdownBar_CounterPlay(self)
	local parent = self:GetParent();
	local w, h = parent:GetSize();
	parent.indicator:SetSize(w, h);
	parent:_dispatch(Counter_OnBegin);
end

function MessCountdownBar_CounterUpdate(self, elapsed)
	local progressLeft = 1 - self.anim:GetSmoothProgress();
	local parent = self:GetParent();
	local w, h = parent:GetSize();
	if parent:GetOrientation() == ORI_VERTICAL then
		parent.indicator:SetSize(w, h * progressLeft);
	else
		parent.indicator:SetSize(w * progressLeft, h);
	end
end

function MessCountdownBar_CounterFinished(self, requested)
	local parent = self:GetParent();
	parent:_dispatch(Counter_OnFinish);
	parent:Hide();
end

function MessCountdownBar_CounterStop(self, requested)
	local parent = self:GetParent();
	parent:Hide();
end

----------------------------------------------------------------	MessHybridScrollBarTemplate
local MessHybridScrollBarMeta = {
	["SetIndicatorColor"] = function(self, color, activeColor)
		color = color or COLOR_INDICATOR;
		activeColor = activeColor or COLOR_INDICATORACTIVED;
		self.indicator:SetVertexColor(unpack(color));
		self.indicator2:SetVertexColor(unpack(color));
		self.indicatorColor = color;
		self.indicatorActiveColor = activeColor;
	end,
	["BeginCountdown"] = function(self, duration)
		if self.countdownTimer:IsPlaying() then	return;	end
		self.countdownTimer.anim:SetDuration(duration or TIMETOCOUNTDOWN);
		self.countdownTimer:Play();
	end,
	["HaltCountdown"] = function(self)
		self.countdownTimer:Stop();
	end,
	["IsCountdown"] = function(self)
		return self.countdownTimer:IsPlaying();
	end,
};

function MessHybridScrollBar_OnLoad(self)
	MessEventDispatcher_OnLoad(self);
	MessScrollBar_OnLoad(self);
	inherits(self, MessHybridScrollBarMeta);
end

function MessHybridScrollBar_OnHide(self)
	self.countdownTimer:Stop();
end

function MessHybridScrollBar_CounterPlay(self)
	local parent = self:GetParent();
	parent.indicator:Hide();
	local w, h = parent:GetSize();
	parent.indicator2:SetSize(w - 2, h - 2);
	parent.indicator2:SetVertexColor(unpack(parent.indicatorColor));
	parent.indicator2:Show();
	parent:_dispatch(Counter_OnBegin);
end

function MessHybridScrollBar_CounterUpdate(self, elapsed)
	local progressLeft = 1 - self.anim:GetSmoothProgress();
	local parent = self:GetParent();
	local w, h = parent:GetSize();
	if parent:GetOrientation() == ORI_VERTICAL then
		parent.indicator2:SetSize(w - 2, (h - 2) * progressLeft);
	else
		parent.indicator2:SetSize((w - 2) * progressLeft, h - 2);
	end
end

function MessHybridScrollBar_CounterFinished(self, requested)
	local parent = self:GetParent();
	parent:_dispatch(Counter_OnFinish);
	parent.indicator2:Hide();
	parent.indicator:Show();
	parent:Update();
end

function MessHybridScrollBar_CounterStop(self, requested)
	local parent = self:GetParent();
	parent.indicator2:Hide();
	parent.indicator:Show();
	parent:Update();
end

----------------------------------------------------------------	MessCompactScrollFrameTemplate
local MessCompactScrollFrameMeta = {
	--	attributes
	["_elapsed"] = 0,
	["_smoothVelocity"] = 0,
	["_smoothTimer"] = nil,
	["scrollStep"] = SCROLLSTEP,	
	--	methods
	["SetScrollStep"] = function(self, step)
		self.scrollStep = step or SCROLLSTEP;
	end,
	["GetScrollStep"] = function(self)
		return self.scrollStep;
	end,
	["VerticalSmoothDeltaScroll"] = function(self, delta, duration)
		duration = duration or SCROLLTIME;
		if not delta then
			--	break & stop
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		elseif not self._smoothTimer then
			--	init scrolling
			self._smoothVelocity = -delta * self.scrollStep / duration;
			self._smoothTimer = duration;
		elseif delta > 0 and 0 > self._smoothVelocity then
			--	accelerate upward
			self._smoothVelocity = self._smoothVelocity - self.scrollStep / self._smoothTimer;
		elseif 0 > delta and self._smoothVelocity > 0 then
			--	accelerate downward
			self._smoothVelocity = self._smoothVelocity + self.scrollStep / self._smoothTimer;
		else
			--	break & stop
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		end
	end,
	["VerticalSmoothScrollTo"] = function(self, offset, duration)
		--	break & re-accelerate
		duration = duration or SCROLLTIME;
		local value = self:GetVerticalScroll();
		local range = self:GetVerticalScrollRange();

		offset = mathBound(offset, 0, range);
		self._smoothVelocity = (offset - value) / duration;
		self._smoothTimer = duration;
	end,
	["SetScrollValue"] = function(self, value, orientation)
		if not orientation or orientation == ORI_VERTICAL then
			self:SetVerticalScroll(value);
		else
			self:SetHorizontalScroll(value);
		end
	end,
	["GetScrollValue"] = function(self, orientation)
		if not orientation or orientation == ORI_VERTICAL then
			return self:GetVerticalScroll();
		else
			return self:GetHorizontalScroll();
		end
	end,
	["GetScrollRange"] = function(self, orientation)
		if not orientation or orientation == ORI_VERTICAL then
			return self:GetVerticalScrollRange();
		else
			return self:GetHorizontalScrollRange();
		end
	end,
	["StackPush"] = function(self, frame)
		self.stack = self.stack or {};
	
		local h = frame:GetHeight();
		frame:ClearAllPoints();
		frame:SetParent(self.content);
		if self.stackPtr then
			frame:SetPoint("TOPLEFT", self.stackPtr, "BOTTOMLEFT");
			frame:SetPoint("TOPRIGHT", self.stackPtr, "BOTTOMRIGHT");
		else
			frame:SetPoint("TOPLEFT", self.content, "TOPLEFT");
			frame:SetPoint("TOPRIGHT", self.content, "TOPRIGHT");
		end
		tinsert(self.stack, frame);
		self.content:SetHeight(self.content:GetHeight() + h);
		self.stackPtr = frame;
	end,
};

function MessCompactScrollFrame_OnLoad(self)
	inherits(self, MessCompactScrollFrameMeta);
	
	self.bar:SetScrollParent(self);
	self.bar:SetFading(true);
end

function MessCompactScrollFrame_OnScrollRangeChanged(self, xrange, yrange)
	self.bar:SetMinMaxValues(0, yrange);
end

function MessCompactScrollFrame_OnVerticalScroll(self, offset)
	self.bar:SetValue(offset);
end

function MessCompactScrollFrame_OnSizeChanged(self, w, h)
	self.content:SetWidth(w);
	self.bar:SetViewportSize(w, h);
end

function MessCompactScrollFrame_OnMouseWheel(self, delta)
	self:VerticalSmoothDeltaScroll(delta);
end

function MessCompactScrollFrame_OnShow(self)
	self._elapsed = 0;
end

function MessCompactScrollFrame_OnUpdate(self, elapsed)
	self._elapsed = self._elapsed + elapsed;
	if THRESHOLD > self._elapsed then return end
	elapsed = mathFloor(self._elapsed / THRESHOLD) * THRESHOLD;
	self._elapsed = self._elapsed - elapsed;
	-- smooth scroll
	if self._smoothTimer then
		local value = self:GetVerticalScroll();
		local range = self:GetVerticalScrollRange();

		value = value + mathMin(elapsed, self._smoothTimer) * self._smoothVelocity;
		self:SetVerticalScroll(mathBound(value, 0, range));

		self._smoothTimer = self._smoothTimer - elapsed;
		if 0 >= self._smoothTimer then
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		end
	end
end

----------------------------------------------------------------	MessFauxScrollFrameTemplate
--	private usages
local function MessFauxScrollFrame_BuildTemplates(self, template, type, gapOffset)
	type = type or "Frame";
	gapOffset = gapOffset or 0;
	local nameFormat = self:GetName() .. "Button%d";
	
	local height;
	if not self.items then
		self.items = {};
		local sample = CreateFrame(type, string.format(nameFormat, 1), self.content, template);
		height = sample:GetHeight();
		sample:ClearAllPoints();
		sample:SetPoint("TOPLEFT");
		sample:SetPoint("RIGHT");
		tinsert(self.items, sample);
	else
		height = self.items[1]:GetHeight();
	end
	
	local unitHeight = mathRound(height) - gapOffset;
	self.unitHeight = unitHeight;
	
	local maxItemCount = mathCeil(self:GetHeight() / unitHeight) + 2;
	for i = #self.items + 1, maxItemCount do
		local item = CreateFrame(type, string.format(nameFormat, i), self.content, template);
		item:ClearAllPoints();
		item:SetPoint("TOPLEFT", self.items[i - 1], "BOTTOMLEFT", 0, gapOffset);
		item:SetPoint("RIGHT");
		tinsert(self.items, item);
	end
	
	self.content:SetWidth(self:GetWidth());
	self.content:SetHeight(maxItemCount * unitHeight);
	
	return self.items;
end

--	translates logical value to scroll offset, also implements <OnVerticalScroll> handle
local function MessFauxScrollFrame_SetOffset(self, value)
	value = mathBound(value, 0, self._range);
	self._offset = value;
	--	faux-ize : index-th top most item in viewport, with offset pixels vertical shift.
	local index, scrollOffset = self:GetOffsetItem(value);
	index = index or 1;
	scrollOffset = scrollOffset or 0;
	--	update faux display
	self:SetVerticalScroll(scrollOffset);
	if self._topItemIndex ~= index then
		self._topItemIndex = index;
		self:Update();
	end
	--	OnVerticalScroll handle
	self.bar:SetValue(value);
end

--	updates content scroll range, also implements <OnScrollRangeChanged> handle
local function MessFauxScrollFrame_UpdateContent(self, totalHeight, viewportHeight)
	self._totalHeight = totalHeight;
	local range = mathMax(0, mathRound(totalHeight - viewportHeight));
	if range ~= self._range then
		self._range = range;
		self.bar:SetMinMaxValues(0, range);
	end
	-- legitimate current offset value
	MessFauxScrollFrame_SetOffset(self, self._offset);
end

local MessFauxScrollFrameMeta = {
	--	attributes
	["_elapsed"] = 0,
	["_smoothVelocity"] = 0,
	["_smoothTimer"] = nil,
	["_totalHeight"] = 0,		--	logical content height
	["_range"] = 0,				--	logical scroll range
	["_offset"] = 0,			--	logical scroll offset
	["_topItemIndex"] = 1,		--	cache index of item which is the top most in viewport
	["scrollStep"] = SCROLLSTEP,
	["unitHeight"] = SCROLLSTEP,
	--	functions
	["GetOffsetItem"] = function(self, offsetPixel)	--	inheritant overwrites <GetOffsetItem>, in case of item with different height
		local index = mathCeil(offsetPixel / self.unitHeight);
		local scrollOffset = offsetPixel - index * self.unitHeight + self.unitHeight;
		return index, mathMax(0, scrollOffset);
	end,
	["Update"] = function(self)						--	inheritant overwrites <Update> to update item displaying.
		if not self:IsVisible() then return; end
		local offsetIndex = self.scrollFrame:GetTopItemIndex();
		for i = 1, #self.scrollFrame.items do
			local index = offsetIndex + i - 1;
			print("data", index, "is in viewport");
		end
	end,
	["Init"] = function(self, template, type, gapOffset)	--	calls MessFauxScrollFrame_BuildTemplates, which builds template items
		local rtn = MessFauxScrollFrame_BuildTemplates(self, template, type, gapOffset);
		self:Update();
		return rtn;
	end,
	["GetTopItemIndex"] = function(self)
		return self._topItemIndex;
	end,
	["SetScrollStep"] = function(self, step)
		self.scrollStep = step or SCROLLSTEP;
	end,
	["GetScrollStep"] = function(self)
		return self.scrollStep;
	end,
	["VerticalSmoothDeltaScroll"] = function(self, delta, duration)
		duration = duration or SCROLLTIME;
		if not delta then
			--	break & stop
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		elseif not self._smoothTimer then
			self._smoothVelocity = -delta * self.scrollStep / duration;
			self._smoothTimer = duration;
		elseif delta > 0 and 0 > self._smoothVelocity then
			--	accelerate upward
			self._smoothVelocity = self._smoothVelocity - self.scrollStep / self._smoothTimer;
		elseif 0 > delta and self._smoothVelocity > 0 then
			--	accelerate downward
			self._smoothVelocity = self._smoothVelocity + self.scrollStep / self._smoothTimer;
		else
			--	break & stop
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		end
	end,
	["VerticalSmoothScrollTo"] = function(self, offset, duration)
		--	break & re-accelerate
		duration = duration or SCROLLTIME;
		self._smoothVelocity = (offset - self._offset) / duration;
		self._smoothTimer = duration;
	end,
	["SetScrollValue"] = function(self, value, orientation)
		if not orientation or orientation == ORI_VERTICAL then
			MessFauxScrollFrame_SetOffset(self, value);
		else
			--
		end
	end,
	["GetScrollValue"] = function(self, orientation)
		if not orientation or orientation == ORI_VERTICAL then
			return self._offset;
		else
			return 0;
		end
	end,
	["SetScrollRange"] = function(self, range, orientation)
		if not orientation or orientation == ORI_VERTICAL then
			MessFauxScrollFrame_UpdateContent(self, range, self:GetHeight());
			self:Update();
		else
		end
	end,
	["GetScrollRange"] = function(self)
		if not orientation or orientation == ORI_VERTICAL then
			return self._range;
		else
			return 0;
		end
	end,
};

function MessFauxScrollFrame_OnLoad(self)
	inherits(self, MessFauxScrollFrameMeta);

	--	scrollbar
	self.bar:SetScrollParent(self);
	self.bar:SetFading(true);
end

function MessFauxScrollFrame_OnSizeChanged(self, w, h)
	self.content:SetSize(w, h);	--necessary
	self.bar:SetViewportSize(w, h);
	MessFauxScrollFrame_UpdateContent(self, self._totalHeight, h);
	self:Update();
end

function MessFauxScrollFrame_OnMouseWheel(self, delta)
	self:VerticalSmoothDeltaScroll(delta);
end

function MessFauxScrollFrame_OnShow(self)
	self._elapsed = 0;
end

function MessFauxScrollFrame_OnUpdate(self, elapsed)
	self._elapsed = self._elapsed + elapsed;
	if THRESHOLD > self._elapsed then return end
	elapsed = mathFloor(self._elapsed / THRESHOLD) * THRESHOLD;
	self._elapsed = self._elapsed - elapsed;
	-- smooth scroll
	if self._smoothTimer then
		local value = self._offset;
		value = value + mathMin(elapsed, self._smoothTimer) * self._smoothVelocity;
		MessFauxScrollFrame_SetOffset(self, value);

		self._smoothTimer = self._smoothTimer - elapsed;
		if 0 >= self._smoothTimer then
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		end
	end
end

----------------------------------------------------------------