--ANCHOR_CURSOR	--	already defined
ANCHOR_RELATIVE = "ANCHOR_RELATIVE";
ANCHOR_MANUAL = "ANCHOR_MANUAL";

local NUM_MAXLETTER = 1024;
local PATHSTR = "INTERFACE\\ICONS\\%s";
local PATHPREFIX = "INTERFACE\\ICONS\\";
local DEFAULT = "INV_MISC_QUESTIONMARK";
local SIZE_BLOCK_W = 60;
local SIZE_BLOCK_H = 48;
local SIZE_GAP = 2;

local COLOR_BG = {0.196, 0.196, 0.196};
local COLOR_BG_HIGHLIGHTEN = {1, 0.5, 0};
local COLOR_BG_DISABLED = {0.14, 0.14, 0.14};
local COLOR_FG = {1, 0.5, 0};
local COLOR_FG_HIGHLIGHTEN = {0.1, 0.1, 0.1};
local COLOR_FG_DISABLED = {0.5, 0.5, 0.5};

local LDB = LibStub("LibDataBroker-1.1");
assert(LDB, "BrokerBrowser : LibDataBroker Missing");

local _iconCache;
--local _borkerCache;	--	broker may be created dynamically (i.e. broker_actionmenu), thus do not cache it.

----------------------------------------------------------------
local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

local function getMacroIcons(cache)	--	index : name (without path)
	cache = cache or {};
	--	character spellbook icons
	cache["CharSpellBook"] = {};
	for i = 1, GetNumSpellTabs() do
		local tab, tabTex, offset, numSpells = GetSpellTabInfo(i);
		offset = offset + 1;
		local tabEnd = offset + numSpells - 1;
		for j = offset, tabEnd do
			local spellType, ID = GetSpellBookItemInfo(j, "player");
			if spellType ~= "FUTURESPELL" then
				local spellTexture = strupper(GetSpellBookItemTexture(j, "player"));
				if not string.match(spellTexture, "INTERFACE\\BUTTONS\\") then
					spellTexture = gsub(spellTexture, PATHPREFIX, "");
					tinsert(cache["CharSpellBook"], spellTexture);
				end
			end
			if (spellType == "FLYOUT") then
				local _, _, numSlots, isKnown = GetFlyoutInfo(ID);
				if (isKnown and numSlots > 0) then
					for k = 1, numSlots do
						local spellID, overrideSpellID, isKnown = GetFlyoutSlotInfo(ID, k)
						if isKnown then
							local spellTexture = gsub(strupper(GetSpellTexture(spellID)), PATHPREFIX, "");
							tinsert(cache["CharSpellBook"], spellTexture);
						end
					end
				end
			end
		end
	end
	--	other icons
	cache["Macro"] = GetMacroIcons({});
	for index, name in ipairs(cache["Macro"]) do
		cache["Macro"][index] = strupper(name);
	end
	cache["MacroItem"] = GetMacroItemIcons({});
	for index, name in ipairs(cache["Macro"]) do
		cache["MacroItem"][index] = strupper(name);
	end

	return cache;
end

local function showAlongAnchor(self, anchor_type, ...)
	self:ClearAllPoints();
	if anchor_type == ANCHOR_CURSOR then
		local offsetX, offsetY = ...;
		local sw, sh = GetScreenWidth(), GetScreenHeight();
		local cursorX, cursorY = GetCursorPosition();
		offsetX = offsetX or 0;
		offsetY = offsetY or 0;
		self:SetPoint(
			string.format("%s%s", cursorY >= sh / 2 and "TOP" or "BOTTOM", cursorX >= sw / 2 and "RIGHT" or "LEFT"),
			UIParent, "BOTTOMLEFT", cursorX - UIParent:GetLeft(), cursorY - UIParent:GetBottom(),
			cursorX >= sw / 2 and -offsetX or offsetX,
			cursorY >= sh / 2 and -offsetY or offsetY
		);
		self:Show();
		self:Raise();
	elseif anchor_type == ANCHOR_RELATIVE then
		local relative, offsetX, offsetY = ...;
		local sw, sh = GetScreenWidth(), GetScreenHeight();
		local x, y, w, h = relative:GetLeft(), relative:GetBottom(), relative:GetSize();
		local centerX, centerY = x + w / 2, y + h / 2;
		offsetX = math.abs(offsetX or 0);
		offsetY = math.abs(offsetY or 0);
		self:SetPoint(
			string.format("%s%s", centerY >= sh / 2 and "TOP" or "BOTTOM", centerX >= sw / 2 and "RIGHT" or "LEFT"),
			relative,
			string.format("%s%s", centerY >= sh / 2 and "TOP" or "BOTTOM", centerX >= sw / 2 and "LEFT" or "RIGHT"),
			centerX >= sw / 2 and -offsetX or offsetX,
			centerY >= sh / 2 and -offsetY or offsetY
		);
		self:Show();
		self:Raise();
	elseif anchor_type == ANCHOR_MANUAL then
		self:SetPoint(...);
		self:Show();
		self:Raise();
	end
end

local function getBrokerObjects(cache)
	cache = cache or {};	--	type : { index : broker obj }
	for name, obj in LDB:DataObjectIterator() do
		if name and obj then
			obj.name = name;
			local type = obj.type or "unknown";
			cache[type] = cache[type] or {};
			tinsert(cache[type], obj);
		end
	end
	return cache;
end

----------------------------------------------------------------
local MessStandaloneEditorMeta = {
	["BeginEditing"] = nil,
	["CancelEditing"] = nil,
	["SetEditable"] = nil,
	["IsEditable"] = nil,
};

function MessStandaloneEditor_OnLoad(self)
	MessContainer_OnLoad(self);
	--inherits(self, MessStandaloneEditorMeta);	--	since its virtual
	self:HookKeyInput("ESCAPE", function(self, key)
		self:CancelEditing();
	end);
end

----------------------------------------------------------------	MessTextEditorTemplate
local MessTextEditorMeta = {
	["BeginEditing"] = function(self, title, text, acceptFunc, cancelFunc, ...)
		self.title:SetText(title);
		self.textfield:BeginEditing(text);
		self.acceptFunc = acceptFunc;
		self.cancelFunc = cancelFunc;

		showAlongAnchor(self, ...);
	end,
	["CancelEditing"] = function(self)
		if not self:IsShown() then return; end
		pcall(self.cancelFunc);
		self.acceptFunc = nil;
		self.cancelFunc = nil;
		self:Hide();
	end,
	["SetEditable"] = function(self, flag)
		self.textfield:SetEditable(flag);
	end,
	["IsEditable"] = function(self)
		return self.textfield:IsEditable();
	end,
	["TextEdit_Modified"] = function(self, textfield, success)
		pcall(success and self.acceptFunc or self.cancelFunc, success and self.textfield:GetText());
		self.acceptFunc = nil;
		self.cancelFunc = nil;
		self:Hide();
	end,
};

function MessTextEditor_OnLoad(self)
	--	inherits
	MessStandaloneEditor_OnLoad(self);
	inherits(self, MessTextEditorMeta);
	--
	self:EnableCloser(false);
	self:EnableSizer(false);

	self.textfield:RegisterCallback(self, "TextEdit_Modified");
end

----------------------------------------------------------------	MessIconEditorTemplate
local MessIconEditorMeta = {
	["_doAccept"] = function(self)
		if not self:IsShown() then return; end
		self:Hide();
		local name = self.massSelector:GetSelection();
		pcall(self.acceptFunc, PATHSTR:format(name or DEFAULT));
		self.acceptFunc = nil;
		self.cancelFunc = nil;
	end,
	["_doCancel"] = function(self)
		if not self:IsShown() then return; end
		self:Hide();
		pcall(self.cancelFunc);
		self.acceptFunc = nil;
		self.cancelFunc = nil;
	end,
	["BeginEditing"] = function(self, name, texture, acceptFunc, cancelFunc, ...)
		if not self.initialized then
			_iconCache = _iconCache or getMacroIcons();
			self.massSelector:MountMassData("CharSpellBook", _iconCache["CharSpellBook"], true);
			self.massSelector:MountMassData("Macro", _iconCache["Macro"], true);
			self.massSelector:MountMassData("MacroItem", _iconCache["MacroItem"]);
			self.initialized = true;
		end
		self.title:SetText(name or "IconBrowser");
		self.acceptFunc = acceptFunc;
		self.cancelFunc = cancelFunc;

		showAlongAnchor(self, ...);
		self:SetSelection(texture);
	end,
	["CancelEditing"] = function(self)
		self:_doCancel();
	end,
	["SetEditable"] = function(self, flag)
		self.accept:SetEnabled(flag);
	end,
	["IsEditable"] = function(self)
		return self.accept:IsEnabled();
	end,
	["SetSelection"] = function(self, texture, noScroll)
		local name = texture and gsub(strupper(texture), PATHPREFIX, "") or DEFAULT;
		self.textureName:SetText(name);
		self.iconBtn:SetIcon(texture or PATHSTR:format(DEFAULT));
		self.massSelector:AddSelection(name);
		if not noScroll then
			self.massSelector:ScrollToData(name);
		end
	end,
	["SetPreferedSize"] = function(self, row, col)
		local w, h = self.massSelector:SetPreferedSize(row, col);
		w = w + 10 + 10 + 10;
		h = h + 10 + 36 + 5 + 10 + 24 + 10;
		self:SetSize(w, h);
	end,
	["Selector_OnSelected"] = function(self, trigger, name, motion)
		if not motion then return; end
		self.textureName:SetText(name);
		self.iconBtn:SetIcon(PATHSTR:format(name));
	end,
};

local function gridIconElementVisualize(massSelector, element, data)
	element.icon:SetTexture(PATHSTR:format(data));
	element:Show();
end

function MessIconEditor_OnLoad(self)
	--	inherits
	MessStandaloneEditor_OnLoad(self);
	inherits(self, MessIconEditorMeta);
	--
	self:EnableMover(false);
	self:EnableCloser(false);
	self:EnableSizer(true);
	--
	self.iconBtn:SetIcon();
	self.iconBtn:SetScript("OnClick", function(btn, button)
		btn:GetParent():SetSelection(btn:GetIcon());
	end);
	self.accept:SetText("Accept", true);
	self.accept:SetScript("OnClick", function(btn, button)
		btn:GetParent():_doAccept();
	end);
	self.cancel:SetText("Cancel", true);
	self.cancel:SetScript("OnClick", function(btn, button)
		btn:GetParent():_doCancel();
	end);
	self.massSelector.Visualize = gridIconElementVisualize;
	self.massSelector:RegisterCallback(self, "Selector_OnSelected");
end

----------------------------------------------------------------	MessCompactIconEditorTemplate
local MessCompactIconEditorMeta = {
	["BeginEditing"] = function(self, texture, acceptFunc, cancelFunc, ...)
		if not self.initialized then
			_iconCache = _iconCache or getMacroIcons();
			self.massSelector:MountMassData("CharSpellBook", _iconCache["CharSpellBook"], true);
			self.massSelector:MountMassData("Macro", _iconCache["Macro"], true);
			self.massSelector:MountMassData("MacroItem", _iconCache["MacroItem"]);
			self.initialized = true;
		end
		self.acceptFunc = acceptFunc;
		self.cancelFunc = cancelFunc;

		showAlongAnchor(self, ...);
		self:SetSelection(texture);
	end,
	["CancelEditing"] = function(self)
		if not self:IsShown() then return; end
		self:Hide();
		pcall(self.cancelFunc);
		self.acceptFunc = nil;
		self.cancelFunc = nil;
	end,
	["SetEditable"] = function(self, flag)
		self.massSelector:SetSelectLimit(flag and 1 or 0);
	end,
	["IsEditable"] = function(self)
		return self.massSelector:GetSelectLimit() ~= 0;
	end,
	["SetSelection"] = function(self, texture, noScroll)
		local name = texture and gsub(strupper(texture), PATHPREFIX, "") or DEFAULT;
		self.massSelector:AddSelection(name);
		if not noScroll then
			self.massSelector:ScrollToData(name);
		end
	end,
	["SetPreferedSize"] = function(self, row, col)
		local w, h = self.massSelector:SetPreferedSize(row, col);
		w = w + 5 + 5;
		h = h + 5 + 15;
		self:SetSize(w, h);
	end,
	["Selector_OnSelected"] = function(self, trigger, name, motion)
		if not self:IsShown() or not motion then return; end
		self:Hide();
		pcall(self.acceptFunc, PATHSTR:format(name), motion);
		self.acceptFunc = nil;
		self.cancelFunc = nil;
	end,
	["Counter_OnFinish"] = function(self, trigger)
		self:CancelEditing();
	end,
	["Frame_OnEnter"] = function(self, trigger)
		self.massSelector.bar:HaltCountdown();
	end,
	["Frame_OnLeave"] = function(self, trigger)
		self.massSelector.bar:BeginCountdown();
	end,
};

local function compactIconElementVisualize(massSelector, element, data)
	element.icon:SetTexture(PATHSTR:format(data));
	element:Show();
end

function MessCompactIconEditor_OnLoad(self)
	--	inherits
	MessStandaloneEditor_OnLoad(self);
	inherits(self, MessCompactIconEditorMeta);
	--
	self:EnableMover(false);
	self:EnableCloser(false);
	self:EnableSizer(false);
	--
	self.massSelector.Visualize = compactIconElementVisualize;
	self.massSelector:RegisterCallback(self, "Selector_OnSelected");
	self.massSelector.bar:RegisterCallback(self, "Counter_OnFinish");
	MessCursorTracer:RegisterTracer(self);
end

----------------------------------------------------------------	MessBrokerElementTemplate
function MessBrokerElement_OnLoad(self)
	self:RegisterForClicks("AnyDown");
	self.bg:SetVertexColor(unpack(COLOR_BG));
	self.text:SetTextColor(unpack(COLOR_FG));
end

function MessBrokerElement_OnEnter(self, motion)
	self.highlight:Show();
	self.highlightAnime:Play();
end

function MessBrokerElement_OnLeave(self, motion)
	self.highlight:Hide();
	self.highlightAnime:Stop();
end

local function messBrokerElementVisualize(massSelector, element, data)
	local shortName = gsub(strupper(data.name), "BROKER_", "");
	shortName = gsub(shortName, "_", " ");
	element.icon:SetTexture(data.icon);
	element.text:SetText(shortName);
	--	exclusions
	local comapctBorkerEditor = massSelector:GetParent();
	if comapctBorkerEditor.exclusionMap[data.name] then
		element.bg:SetVertexColor(unpack(COLOR_BG_DISABLED));
		element.icon:SetDesaturated(true);
		element.text:SetTextColor(unpack(COLOR_FG_DISABLED));
		element:SetEnabled(false);
		print("debug:", name, "matched");
	else
		element.bg:SetVertexColor(unpack(COLOR_BG));
		element.icon:SetDesaturated(false);
		element.text:SetTextColor(unpack(COLOR_FG));
		element:SetEnabled(true);
		--print("debug:", name, "no match");
	end
	element:Show();
end

----------------------------------------------------------------	MessCompactBrokerEditorTemplate
local MessCompactBrokerEditorMeta = {
	["exclusionMap"] = nil,	--	name : excluded broker name
	["BeginEditing"] = function(self, excludeList, acceptFunc, cancelFunc, ...)
		if not self.initialized then
			for type, list in pairs(getBrokerObjects()) do
				self.massSelector:MountMassData(type, list, true);
			end
			self.initialized = true;
		end
		if excludeList then
			wipe(self.exclusionMap);
			for _, name in ipairs(excludeList) do
				self.exclusionMap[name] = true;
				print("debug:", name, "existed");
			end
		end
		self.acceptFunc = acceptFunc;
		self.cancelFunc = cancelFunc;

		showAlongAnchor(self, ...);
	end,
	["CancelEditing"] = function(self)
		if not self:IsShown() then return; end
		self:Hide();
		pcall(self.cancelFunc);
		wipe(self.exclusionMap);
		self.acceptFunc = nil;
		self.cancelFunc = nil;
	end,
	["SetEditable"] = function(self, flag)
		self.massSelector:SetSelectLimit(flag and -1 or 0);
	end,
	["IsEditable"] = function(self)
		return self.massSelector:GetSelectLimit() ~= 0;
	end,
	["SetPreferedSize"] = function(self, row, col)
		local w, h = self.massSelector:SetPreferedSize(row, col);
		w = w + 5 + 5;
		h = h + 5 + 15;
		self:SetSize(w, h);
	end,
	["Selector_OnSelected"] = function(self, trigger, data, motion)
		if not self:IsShown() or not motion then return; end
		self:Hide();
		pcall(self.acceptFunc, data, motion);
		wipe(self.exclusionMap);
		self.acceptFunc = nil;
		self.cancelFunc = nil;
	end,
	["Counter_OnFinish"] = function(self, trigger)
		self:CancelEditing();
	end,
	["Frame_OnEnter"] = function(self, trigger)
		self.massSelector.bar:HaltCountdown();
	end,
	["Frame_OnLeave"] = function(self, trigger)
		self.massSelector.bar:BeginCountdown();
	end,
};

function MessCompactBrokerEditor_OnLoad(self)
	--	inherits
	MessStandaloneEditor_OnLoad(self);
	inherits(self, MessCompactBrokerEditorMeta);
	--
	self.exclusionMap = {};
	self:EnableMover(false);
	self:EnableCloser(false);
	self:EnableSizer(false);
	MessCursorTracer:RegisterTracer(self);
	--	massSelector
	self.massSelector.elementWidth = SIZE_BLOCK_W;
	self.massSelector.elementHeight = SIZE_BLOCK_H;
	self.massSelector:SetTemplate("MessBrokerElementTemplate");
	self.massSelector.elementGap = SIZE_GAP;
	self.massSelector:SetSelectLimit(-1);	--	unlimited selection
	self.massSelector.Visualize = messBrokerElementVisualize;
	self.massSelector:RegisterCallback(self, "Selector_OnSelected");
	self.massSelector.bar:RegisterCallback(self, "Counter_OnFinish");
end

----------------------------------------------------------------	MessMacroEditorTemplate
local MessMacroEditorMeta = {
	["_clearFocus"] = function(self)
		self.iconSelector:CancelEditing();
		self:SetLock(false);
		self.nameInput:ClearFocus();
		self.textScroller:ClearFocus();
	end,
	["_doAccept"] = function(self)
		if not self:IsShown() then return; end
		self:_clearFocus();
		self:Hide();
		pcall(self.acceptFunc, self.iconBtn:GetIcon(), self.nameInput:GetText(), self.textScroller:GetText());
		self.acceptFunc = nil;
		self.cancelFunc = nil;
	end,
	["_doCancel"] = function(self)
		if not self:IsShown() then return; end
		self:_clearFocus();
		self:Hide();
		pcall(self.cancelFunc);
		self.acceptFunc = nil;
		self.cancelFunc = nil;
	end,
	["BeginEditing"] = function(self, title, texture, name, body, acceptFunc, cancelFunc, ...)
		self.title:SetText(title);
		self.iconBtn:SetIcon(texture);
		self.nameInput:BeginEditing(name);
		self.textScroller:SetText(body);
		self.acceptFunc = acceptFunc;
		self.cancelFunc = cancelFunc;

		showAlongAnchor(self, ...);
	end,
	["CancelEditing"] = function(self)
		self:_doCancel();
	end,
	["SetEditable"] = function(self, flag)
		self.iconBtn:SetEnabled(flag);
		self.nameInput:SetEditable(flag);
		self.textScroller:SetEditable(flag);
		self.save:SetEnabled(flag);
	end,
	["IsEditable"] = function(self)
		return self.save:IsEnabled();
	end,
	["SetMaxTextLetter"] = function(self, number)
		self.textScroller:SetMaxLetters(number);
	end,
};

function MessMacroEditor_OnLoad(self)
	--	inherits
	MessStandaloneEditor_OnLoad(self);
	inherits(self, MessMacroEditorMeta);
	--
	self:EnableCloser(false);
	self:SetMaxTextLetter(NUM_MAXLETTER);
	self.textScroller.textarea:SetScript("OnTextChanged", function(self)
		MessMacroEditor.counter:SetFormattedText("%4d / %4d", self:GetNumLetters(), self:GetMaxLetters());
	end);
	self.iconSelector:SetPreferedSize(4, 6);
	self.iconBtn:SetScript("OnCLick", function(self, button)
		local parent = self:GetParent();
		parent:_clearFocus();
		parent:SetLock(true);
		--	iconSelector
		parent.iconSelector:BeginEditing(self:GetIcon(), function(texture, motion)
				self:SetIcon(texture);
				parent:_clearFocus();
			end, function()
				parent:_clearFocus();
			end, ANCHOR_RELATIVE, self, 10, 5);
	end);
	self.save:SetText("Save", true);
	self.save:SetScript("OnCLick", function(self, button)
		self:GetParent():_doAccept();
	end);
	self.cancel:SetText("Cancel", true);
	self.cancel:SetScript("OnCLick", function(self, button)
		self:GetParent():_doCancel();
	end);
end

