local _name, _addon = ...;

local CONFIG = Broker_ActionMenuConfig;
local MAIN = CONFIG.main;
local ENTRIES = CONFIG.entries;
local OPTIONS = CONFIG.options;

local BASE_TYPE = _addon.BASE_TYPE;
local EXTRA_TYPE = _addon.EXTRA_TYPE;

---------------------------------------------------------------------------------------
function CONFIG:Init()
	self:EnableMover(true);
	self:EnableCloser(true);
	self:EnableSizer(true);
	self:SetTitle(_name);

	self.onResizedFunc = function(self)
		MAIN:RecreateTemplates();
		ENTRIES:RecreateTemplates();
	end;
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
	self:HookKeyInput("ESCAPE", function(self, key)
		if self.panel == MAIN then
			self:Hide();
		elseif self.panel == ENTRIES then
			self:SwitchPanel(MAIN);
		elseif self.panel == OPTIONS then
			self:SwitchPanel(MAIN);
		end
	end);
	self:SetScript("OnEvent", function(self, event, ...)
		local f = self[event];
		if f then
			f(self,...);
		end
	end);
	
	MAIN:Init();
	ENTRIES:Init();
	OPTIONS:Init();
	self.panels = {};
	tinsert(self.panels, MAIN);
	tinsert(self.panels, ENTRIES);
	tinsert(self.panels, OPTIONS);
	
	self.initialized = true;
end

function CONFIG:PLAYER_REGEN_DISABLED()
	self:Hide();
end

function CONFIG:SwitchPanel(panel, skipAnime)
	panel = panel or MAIN;
	for _, p in ipairs(self.panels) do
		if p == panel then
			p:Active(skipAnime);
		else
			p:Deactive(skipAnime);
		end
	end
	self.panel = panel;
end

function CONFIG:GetCurrentPanel()
	return self.panel;
end

function CONFIG:ForceFocus(focus)
	if not focus then return; end
	local cat, index = _addon:Core_IndexOfMenuGroup(focus.data);
	local isMenugroup = cat and index;
	local isEntry = focus.type and (BASE_TYPE[focus.type] or EXTRA_TYPE[focus.type]);
	if isMenugroup then
		self:SwitchPanel(MAIN, true);
		MAIN:SetFocus(focus);
	elseif isEntry then
		self:SwitchPanel(ENTRIES, true);
		ENTRIES:SetFocus(focus);
	end
end

---------------------------------------------------------------------------------------
function _addon:Config_GC(target)
	target = _addon:Core_GC(target);
	local message;
	if not target then
		message = string.format("All garbages of %s are collected.", _name);
	elseif type(target) == "string"  then
		--	kill by specific DB category
		message = string.format("All garbages under %s database are collected.", target);
	else
		--	kill by specific menugroup
		message = string.format("All garbages under %s are collected.", target.data.name);
	end
	MessMessageFrame:ShowMessage(message);
end

function _addon:Config_Show(menugroup)
	if InCombatLockdown() then
		print(_name,": please leave combat");
		return;
	end
	if not CONFIG.initialized then
		CONFIG:Init();
	end
	CONFIG:Show();
	CONFIG:SetLock(false);
	if menugroup then
		CONFIG:SwitchPanel(MAIN, true);
		MAIN:SetFocus(menugroup);
		CONFIG:SwitchPanel(ENTRIES);
	else
		CONFIG:SwitchPanel(MAIN, true);
		MAIN:SetFocus();
	end
end

SLASH_ACTIONMENU1 = "/am";
SlashCmdList.ACTIONMENU = function(msg, editBox)
	_addon:Config_Show();
end
