local _name, _addon = ...;
local _;

local BROKER_TYPE = "launcher";
local CAT_ACCOUNT = "account";
local CAT_CHARACTER = "character";

local LDB = LibStub("LibDataBroker-1.1");
local Dewdrop = LibStub("Dewdrop-2.0");

local TEXT_REMOVED;		--	def it latter
local TEXT_REMOVED_TIP = "to remove this launcher completely : reload ui or manually disable it in broker displayer";
local TOOLTIP_REMOVED;	--	def it latter
local ICON_REMOVED = "Interface\\Icons\\INV_Misc_QuestionMark";
local ICON_DEFAULT = "Interface\\Icons\\INV_Misc_QuestionMark";
local CATEGORY_DEFUALT = "Interface\\addons\\" .. _name .. "\\Medias\\folder-32";

---------------------------------------------------------------------------------	type def
local BASE_TYPE = {
	["COMPANION"] = "companion",
	["BATTLEPET"] = "battlepet",
	["SPELL"] = "spell",
	["ITEM"] = "item",
	["MACRO"] = "macro",
	["EQUIPMENTSET"] = "equipmentset",
};

local EXTRA_TYPE = {
	["ADDONMACRO"] = "addonMacro",
	["LAUNCHER"] = "launcher",
	["SEPARATOR"] = "separator",
	["CATEGORY"] = "category",
};

_addon.BASE_TYPE = BASE_TYPE;
_addon.EXTRA_TYPE = EXTRA_TYPE;

---------------------------------------------------------------------------------	caches
local VALID_ACTION;	--	trigger by ACTIONBAR_SHOWGRID/ACTIONBAR_HIDEGRID
local DB;
local MenuGroupCache = {};	--	data : menugroup
---------------------------------------------------------------------------------
--[[ legacy
local maxDigitCnt = 1;
local formatStr = " %1d";
local function formatDigit(digit)
	digit = digit or 0;
	local exp = ceil(math.log10(tonumber(digit)));
	if exp > maxDigitCnt then
		maxDigitCnt = exp;
		formatStr = " %"..exp.."d";
	end
	return string.format(formatStr, digit);
end

local DAY, HOUR, MINUTE = 86400, 3600, 60;
local function formatTimeRough(sec)
	if not sec or sec < 0 then
		return "";
	end
	local day = floor(sec/DAY);
	local hour = floor(sec/HOUR);
	local minu = floor(sec/MINUTE);
	sec = floor(sec);
	if day > 0 then
		return format(" %2dd", day);
	elseif hour > 0 then
		return format(" %2dh", hour);
	elseif minu > 0 then
		return format(" %2dm", minu);
	else
		return " <1m";
	end
end
]]

local function codeColorString(text, r, g, b, a)
	r = r and (r > 1 and r or r * 255) or 1;
	g = g and (g > 1 and g or g * 255) or 1;
	b = b and (b > 1 and b or b * 255) or 1;
	a = a and (a > 1 and a or a * 255) or 1;
	return string.format("|c%.2X%.2X%.2X%.2X%s%s", a, r, g, b, text, FONT_COLOR_CODE_CLOSE);
end

TEXT_REMOVED = codeColorString("removed", LIGHTYELLOW_FONT_COLOR.r, LIGHTYELLOW_FONT_COLOR.g, LIGHTYELLOW_FONT_COLOR.b);

TOOLTIP_REMOVED = function(tooltip)
	tooltip:AddLine(TEXT_REMOVED_TIP, LIGHTYELLOW_FONT_COLOR.r, LIGHTYELLOW_FONT_COLOR.g, LIGHTYELLOW_FONT_COLOR.b);
end

local function generateBrokerName(prefix)
	local id = 1;
	local name = string.format("%s-%d", prefix, id);
	while LDB:GetDataObjectByName(name) do
		id = id + 1;
		name = string.format("%s_%d", prefix, id);
	end
	return name;
end

local function updateDB()
	Broker_ActionMenuGlobalDB = Broker_ActionMenuGlobalDB or {};
	Broker_ActionMenuGlobalDB.menuGroupList = Broker_ActionMenuGlobalDB.menuGroupList or {};
	Broker_ActionMenuDB = Broker_ActionMenuDB or {};
	Broker_ActionMenuDB.menuGroupList = Broker_ActionMenuDB.menuGroupList or {};

	DB = {};
	DB[CAT_ACCOUNT] = Broker_ActionMenuGlobalDB.menuGroupList;
	DB[CAT_CHARACTER] = Broker_ActionMenuDB.menuGroupList;

	--[[ pre 3.0 competibility
	local version = GetAddOnMetadata(_name, "Version");
	if Broker_ActionMenuGlobalDB.ver ~= version then
		for _, mg in ipairs(DB[CAT_ACCOUNT]) do
			for _, entry in ipairs(mg.subList) do
				entry.subType = nil;
				if not entry.data then
					entry.garbage = true;
				end
			end
		end
	end
	if Broker_ActionMenuDB.ver ~= version then
		for _, mg in ipairs(DB[CAT_CHARACTER]) do
			for _, entry in ipairs(mg.subList) do
				entry.subType = nil;
				if not entry.data then
					entry.garbage = true;
				end
			end
		end
	end
	Broker_ActionMenuGlobalDB.ver = version;
	Broker_ActionMenuDB.ver = version;
	]]
end

---------------------------------------------------------------------------------	creating case
local EntryTemplate = {
	--	common
	["type"] = nil,
	["name"] = nil,
	["texture"] = ICON_DEFAULT,
	["data"] = nil,			--	correspond data, for back-reference
	["cmd"] = nil,			--	macro action command, e.g. /use, /execute, /cast, ...etc
	["isGarbage"] = nil,	--	whether current entry is marked as garbage
	["comment"] = nil,
	--	CATEGORY
	["subList"] = nil,		--	subMenu list > index : entry
	--	MACRO, EQUIPMENTSET
	["char"] = nil,			--	character specific entry (by name)
	["outOfSync"] = nil,	--	if current entry is out-of-sync
	--	LAUNCHER
	["action"] = nil,
	--	root
	["label"] = nil,
	["text"] = nil,
	["showFolderIcon"] = nil,
	["persisted"] = nil,
	["hideMissing"] = nil,
	["hideCooling"] = nil,
};

local entryMeta = {__index = EntryTemplate};

local CASE_CREATOR = {
	[BASE_TYPE["COMPANION"]] = function(...)
		local _, index, subType = ...;
		local entry = setmetatable({}, entryMeta);
		local _, name, spellID, texture = GetCompanionInfo(subType, index);
		entry.type = BASE_TYPE["COMPANION"];
		entry.name = name;
		entry.texture = texture;
		entry.data = spellID;
		entry.cmd = string.format("/cast %s\n", name);
		return entry;
	end,
	[BASE_TYPE["BATTLEPET"]] = function(...)
		local _, petGUID = ...;
		local entry = setmetatable({}, entryMeta);
		local _, customName, _, _, _, _, _, petName, petIcon = C_PetJournal.GetPetInfoByPetID(petGUID);
		entry.type = BASE_TYPE["BATTLEPET"];
		entry.name = customName or petName;
		entry.texture = petIcon;
		entry.data = petGUID;
		return entry;
	end,
	[BASE_TYPE["SPELL"]] = function(...)
		local _, index, spellBook, spellID = ...;
		local entry = setmetatable({}, entryMeta);
		local name, rank, texture = GetSpellInfo(spellID);
		entry.type = BASE_TYPE["SPELL"];
		entry.name = (rank and strlen(rank) > 0) and string.format("%s(%s)", name, rank) or name;
		entry.texture = texture;
		entry.data = spellID;
		entry.cmd = string.format("/cast %s\n", name);
		return entry;
	end,
	[BASE_TYPE["ITEM"]] = function(...)
		local _, itemID, link = ...;
		local entry = setmetatable({}, entryMeta);
		entry.type = BASE_TYPE["ITEM"];
		entry.name = GetItemInfo(itemID);
		entry.texture = GetItemIcon(itemID);
		entry.data = link;
		entry.cmd = "";
		if IsEquippableItem(itemID) then
			entry.cmd = entry.cmd .. string.format("/equip item:%d\n", itemID);
		end
		if PlayerHasToy(itemID) then
			entry.cmd = entry.cmd .. string.format("/use %s\n", entry.name);
		elseif IsUsableItem(itemID) then
			entry.cmd = entry.cmd .. string.format("/use item:%d\n", itemID);
		end
		return entry;
	end,
	[BASE_TYPE["MACRO"]] = function(...)			--	reference by macroID, which changes a lot
		MacroFrameCancelButton_OnClick();	--	cancel current editing, else the ID might changes
		local _, macroID = ...;
		local entry = setmetatable({}, entryMeta);
		entry.type = BASE_TYPE["MACRO"];
		entry.name, entry.texture, entry.cmd = GetMacroInfo(macroID);
		entry.data = macroID;
		entry.char = macroID > MAX_ACCOUNT_MACROS and GetUnitName("player", true);
		return entry;
	end,
	[BASE_TYPE["EQUIPMENTSET"]] = function(...)		--	reference by name
		local _, name = ...;
		local entry = setmetatable({}, entryMeta);
		entry.type = BASE_TYPE["EQUIPMENTSET"];
		entry.name = name;
		--entry.texture = string.format("Interface\\Icons\\%s", GetEquipmentSetInfoByName(name));
		entry.data = name;
		entry.char = GetUnitName("player", true);
		--entry.cmd = string.format("/equipset %s\n", name);
		return entry;
	end,
	[EXTRA_TYPE["ADDONMACRO"]] = function()
		local entry = setmetatable({}, entryMeta);
		entry.type = EXTRA_TYPE["ADDONMACRO"];
		entry.name = "addon macro";
		entry.texture = ICON_DEFAULT;
		entry.cmd = "";
		return entry;
	end,
	[EXTRA_TYPE["LAUNCHER"]] = function(...)		--	reference by name
		local _, name, texture, motion = ...;
		local entry = setmetatable({}, entryMeta);
		entry.type = EXTRA_TYPE["LAUNCHER"];
		entry.name = name;
		entry.texture = texture;
		entry.data = name;
		entry.action = motion or "LeftButton";
		return entry;
	end,
	[EXTRA_TYPE["SEPARATOR"]] = function()
		local entry = {};
		entry.type = EXTRA_TYPE["SEPARATOR"];
		return entry;
	end,
	[EXTRA_TYPE["CATEGORY"]] = function()
		local entry = setmetatable({}, entryMeta);
		entry.type = EXTRA_TYPE["CATEGORY"];
		entry.name = "category";
		entry.texture = CATEGORY_DEFUALT;
		entry.subList = {};
		return entry;
	end,
};

---------------------------------------------------------------------------------	menugroup
local MenuGroup = {
	["data"] = nil,
	["broker"] = nil,
	["_parentMap"] = nil,	--	flat the tree hierarchy > entry : parent entry
};

function MenuGroup:_recursiveMapping(root)
	for index, entry in ipairs(root.subList) do
		self._parentMap[entry] = root;
		if entry.subList then
			self:_recursiveMapping(entry);
		end
	end
end

function MenuGroup:_recursiveGarbaging(root)
	for index = #root.subList, 1, -1 do
		local entry = root.subList[index];
		if entry.isGarbage then
			_addon:Sync_DecacheEntry(entry);
			tremove(root.subList, index);
		elseif entry.subList then
			self:_recursiveGarbaging(entry);
		end
	end
end

function MenuGroup:_createBrokerObj()
	local onClick = function(frame, button)
		local root = self.data;	--	self is menugroup
		if root.isGarbage then
			--	nothing
		elseif VALID_ACTION and GetCursorInfo() then
			local created = self:CreateSubentry(GetCursorInfo());
			if created then
				ClearCursor();
				_addon:Core_Refresh(created);
			end
		elseif IsModifierKeyDown() then
			_addon:Config_Show(self);
		elseif root.subList and #root.subList > 0 then
			Dewdrop:Open(frame, "cursorX", true, "cursorY", true, "children", function(level, value)
				if not Dewdrop:IsOpen(frame) then return; end
				_addon:Menu_Feed(frame, root, level);
			end);
		end
	end
	local onTooltip = function(tooltip)
		if not tooltip then return; end
		local root = self.data;	--	self is menugroup
		if root.comment and strlen(root.comment) > 0 then
			tooltip:AddLine(root.comment);
		end
	end;
	return LDB:NewDataObject(self.data.name, {
		["type"] = BROKER_TYPE,
		["name"] = self.data.name,
		["label"] = self.data.label,
		["text"] = self.data.text,
		["icon"] = self.data.texture,
		["OnClick"] = onClick,
		["OnTooltipShow"] = onTooltip,
	});
end

function MenuGroup:_newMenuGroup(data)
	if not data then
		data = setmetatable({}, entryMeta);
		data.type = EXTRA_TYPE["CATEGORY"];
		local name = generateBrokerName("Menu");
		data.name = name;
		data.label = name .. "_Label";
		data.text = name .. "_Text";
		data.comment = "";
		data.subList = {};
	end

	local newMg = setmetatable({}, {__index = self});
	newMg.data = data;
	newMg.broker = newMg:_createBrokerObj();
	newMg._parentMap = {};
	if newMg.data.subList then
		newMg:_recursiveMapping(newMg.data);
	end

	return newMg;
end

function MenuGroup:_clone(data)
	data = data or self.data;
	local cloneData = CopyTable(data);
	cloneData.name = cloneData.name .. " clone";
	return self:_newMenuGroup(cloneData);
end

function MenuGroup:UpdateBrokerObj(overrideLabel, overrideText, overrideIcon, overrideTip)
	self.broker = self.broker or self:_createBrokerObj();
	self.broker.label = overrideLabel or self.data.label;
	self.broker.text = overrideText or self.data.text;
	self.broker.icon = overrideIcon or self.data.texture;
	if overrideTip then
		self.broker.OnTooltipShow = overrideTip;
	end
end

function MenuGroup:CreateSubentry(...)	--	parent, ... (as last one under parent) / parent, index, ... (as index-th under parent) / index, ... (as index-th under root) / ... (as last one under root)
	local arg1, arg2 = ...;
	local args, parent, index = {};
	if type(arg1) == "table" then
		parent = arg1;
		if type(arg2) == "number" then
			index = arg2;
			for i = 3, select("#", ...) do
				local arg = select(i, ...);
				tinsert(args, arg);
			end
		else
			for i = 2, select("#", ...) do
				local arg = select(i, ...);
				tinsert(args, arg);
			end
		end
	else
		if type(arg1) == "number" then
			index = arg1;
			for i = 2, select("#", ...) do
				local arg = select(i, ...);
				tinsert(args, arg);
			end
		else
			for i = 1, select("#", ...) do
				local arg = select(i, ...);
				tinsert(args, arg);
			end
		end
	end
	local datatype = args[1];
	assert(CASE_CREATOR[datatype], "unsupport base type:"..tostring(datatype));
	local newEntry = CASE_CREATOR[datatype](unpack(args));
	parent = parent or self.data;
	if index then
		tinsert(parent.subList, index, newEntry);
	else
		tinsert(parent.subList, newEntry);
	end
	self._parentMap[newEntry] = parent;
	_addon:Sync_CacheEntry(newEntry);
	return newEntry;
end

function MenuGroup:HasEntry(target)
	return self._parentMap[target] or self.data == target;
end

function MenuGroup:ParentOf(target)
	return self._parentMap[target];
end

function MenuGroup:IndexOf(target)
	local parent = self:ParentOf(target);
	if parent then
		for index, entry in pairs(parent.subList) do
			if entry == target then
				return index;
			end
		end
	end
end

function MenuGroup:Relocate(target, toParent, toIndex)
	local parent, index = self:ParentOf(target), self:IndexOf(target);
	target = tremove(parent.subList, index);
	if toIndex then
		tinsert(toParent.subList, toIndex, target);
	else
		tinsert(toParent.subList, target);
	end
	self._parentMap[target] = toParent;
end

function MenuGroup:GetActiveBrokerNameList()
	local list = {};
	for entry, _ in ipairs(self._parentMap) do
		if entry.type == EXTRA_TYPE["LAUNCHER"] and not entry.isGarbage then
			--local obj = LDB:GetDataObjectByName(entry.data);
			tinsert(list, entry.data);
		end
	end
	return list;
end

function MenuGroup:IsGarbage(data)
	return self:HasEntry(data) and data.isGarbage or self.data.isGarbage;
end

function MenuGroup:CleanGarbage(root)
	if not root then
		if self.data.isGarbage then
			_addon:Core_DeleteMenuGroup(self);
		else
			self:_recursiveGarbaging(self.data);
		end
	else
		self:_recursiveGarbaging(root);
	end
end

---------------------------------------------------------------------------------
local addon = CreateFrame("Frame", nil, UIParent);
addon:UnregisterAllEvents();
addon:Hide();

function addon:ADDON_LOADED(name)
	if name == _name then
		self:UnregisterEvent("ADDON_LOADED");
		updateDB();
		LoadAddOn("Blizzard_MacroUI");
		--LoadAddOn("Blizzard_PetJournal");
		self:RegisterEvent("PLAYER_LOGIN");
	end
end

function addon:PLAYER_LOGIN()
	self:UnregisterEvent("PLAYER_LOGIN");
	
	-- re-bind all menugroups
	for _, data in ipairs(DB[CAT_ACCOUNT]) do
		local mg = MenuGroup:_newMenuGroup(data);
		MenuGroupCache[data] = mg;
		_addon:Sync_CacheEntry(data);
	end
	for _, data in ipairs(DB[CAT_CHARACTER]) do
		local mg = MenuGroup:_newMenuGroup(data);
		MenuGroupCache[data] = mg;
		_addon:Sync_CacheEntry(data);
	end

	self:RegisterEvent("ACTIONBAR_SHOWGRID");
	self:RegisterEvent("ACTIONBAR_HIDEGRID");
	self:RegisterEvent("UPDATE_MACROS");
	self:RegisterEvent("EQUIPMENT_SETS_CHANGED");
	self:RegisterUnitEvent("UNIT_INVENTORY_CHANGED", "player");
	_addon:Sync_HookMacroFuncs();
	if CanUseEquipmentSets() then
		_addon:Sync_HookEquipmentSetFunc();
	end
end

function addon:ACTIONBAR_SHOWGRID()
	VALID_ACTION = true;
end

function addon:ACTIONBAR_HIDEGRID()
	VALID_ACTION = nil;
end

function addon:UPDATE_MACROS()
	_addon:Core_Refresh();
end

function addon:EQUIPMENT_SETS_CHANGED()
	_addon:Core_Refresh();
end

function addon:UNIT_INVENTORY_CHANGED(unitID)
	_addon:Core_Refresh();
end

addon:SetScript("OnEvent", function(self, event, ...)
	local f = self[event];
	if f then
		f(self,...);
	end
end);

addon:RegisterEvent("ADDON_LOADED");

---------------------------------------------------------------------------------
function _addon:Core_IndexOfMenuGroup(menugroup)
	for index, data in ipairs(DB[CAT_ACCOUNT]) do
		if data == menugroup.data then
			return CAT_ACCOUNT, index;
		end
	end
	for index, data in ipairs(DB[CAT_CHARACTER]) do
		if data == menugroup.data then
			return CAT_CHARACTER, index;
		end
	end
end

function _addon:Core_CreateMenuGroup(category, index)
	local mg = MenuGroup:_newMenuGroup();	
	if index then
		tinsert(DB[category], index, mg.data);
	else
		tinsert(DB[category], mg.data);
	end
	MenuGroupCache[mg.data] = mg;
	return mg;
end

function _addon:Core_CloneMenuGroup(menugroup)
	local cloneMG = menugroup:_clone();
	local category, index = self:Core_IndexOfMenuGroup(menugroup);
	tinsert(DB[category], index + 1, cloneMG.data);
	MenuGroupCache[cloneMG.data] = cloneMG;
	return cloneMG;
end

function _addon:Core_DeleteMenuGroup(menugroup)
	local category, index = self:Core_IndexOfMenuGroup(menugroup);
	local data = tremove(DB[category], index);
	menugroup:UpdateBrokerObj(
		codeColorString(data.name, LIGHTYELLOW_FONT_COLOR.r, LIGHTYELLOW_FONT_COLOR.g, LIGHTYELLOW_FONT_COLOR.b),
		TEXT_REMOVED,
		ICON_REMOVED,
		TOOLTIP_REMOVED);
	_addon:Sync_DecacheEntry(data);
	MenuGroupCache[data] = nil;
end
	
function _addon:Core_GetCategoryItemNumber(category)
	return #DB[category];
end

function _addon:Core_GetMenuGroupByIndex(category, index)
	local data = DB[category][index];
	return MenuGroupCache[data];
end
	
function _addon:Core_SwapCategory(menugroup)
	local category, index = self:Core_IndexOfMenuGroup(menugroup);
	local data = tremove(DB[category], index);
	category = category == CAT_ACCOUNT and CAT_CHARACTER or CAT_ACCOUNT;
	tinsert(DB[category], data);
end

function _addon:Core_Refresh(reference)
	if Broker_ActionMenuConfig:IsShown() then
		Broker_ActionMenuConfig:ForceFocus(reference);
	end
	if Dewdrop:IsOpen() then
		Dewdrop:Refresh();
	end
end

function _addon:Core_GC(target)
	if not target then
		--	kill all
		for _, mg in pairs(MenuGroupCache) do
			mg:CleanGarbage();
		end
	elseif type(target) == "string"  then
		--	kill by specific category
		for _, data in pairs(DB[target]) do		--	ipair would act strange on removing-in-the-loop
			MenuGroupCache[data]:CleanGarbage();
		end
	else
		--	kill by specific menugroup
		target:CleanGarbage();
	end
	-- system gc
	collectgarbage();
	return target;
end

---------------------------------------------------------------------------------