local _name, _addon = ...;

local mathFloor = math.floor;

local CONFIG = Broker_ActionMenuConfig;
local MAIN = CONFIG.main;
local ENTRIES = CONFIG.entries;
local OPTIONS = CONFIG.options;
local FloatButtons = FloatButtons_Entry;

local GlassPane = Broker_ActionMenuGlassPane;
local MacroEditor = Broker_ActionMenuMacroEditor;
local IconSelector = Broker_ActionMenuIconSelector;
local BrokerSelector = Broker_ActionMenuBrokerSelector;

local TextEditor = Broker_ActionMenuTextEditor;
local VirtualEntry = ActionMenuVirtualEntry;
local VirtualGapEntry = ActionMenuVirtualGapEntry;

local MEDIA_PATH = "Interface\\addons\\" .. _name .. "\\Medias\\%s";
local ICON_ADD = string.format(MEDIA_PATH, "document_add");
local ICON_MOVE = string.format(MEDIA_PATH, "document_download");

local THRESHOLD = 0.01;

local BASE_TYPE = _addon.BASE_TYPE;
local EXTRA_TYPE = _addon.EXTRA_TYPE;
local HEIGHT_MAP = {
	[EXTRA_TYPE["SEPARATOR"]] = 30,
	[EXTRA_TYPE["CATEGORY"]] = 40,
	["default"] = 40,
};
local SIZE_ICONSMALL = 24;
local SIZE_ICONLARGE = 32;

local LEVEL_OFFSET = 15;
local UPPERBOUND = 0.2;
local LOWERBOUND = 0.8;
local MEDIA_PATH = "Interface\\addons\\" .. _name .. "\\Medias\\%s";
local STR_TYPES = {
	[BASE_TYPE["COMPANION"]] = "Companion",
	[BASE_TYPE["BATTLEPET"]] = "Battlepet",
	[BASE_TYPE["SPELL"]] = "Spell",
	[BASE_TYPE["ITEM"]] = "Item",
	[BASE_TYPE["MACRO"]] = "Macro",
	[BASE_TYPE["EQUIPMENTSET"]] = "Equipment Set",
	[EXTRA_TYPE["ADDONMACRO"]] = "Addon Macro",
	[EXTRA_TYPE["LAUNCHER"]] = "Broker Launcher",
};

local COLOR_BG = {0.196, 0.196, 0.196};
local COLOR_BG_HIGHLIGHTEN = {1, 0.5, 0};
local COLOR_BG_DISABLED = {0.14, 0.14, 0.14};
local COLOR_FG = {1, 0.5, 0};
local COLOR_FG_HIGHLIGHTEN = {0.1, 0.1, 0.1};
local COLOR_FG_DISABLED = {0.5, 0.5, 0.5};
local COLOR_FLOATBUTTON = {0.17, 0.135, 0.118};
local COLOR_FLOATBUTTON_DISABLE = {0.337, 0.267, 0.231};
local COLOR_FLOATBUTTON_DISABLE_ALTER = {0.847, 0.451, 0};
local COLOR_FLOATTIP = {0.627, 0.588, 0.573};

local ALPHA_DRAGGED = 0.25;

local MODE_NORMAL = "normal";
local MODE_DROPIN = "dropping in";
local MODE_DRAGOUT_INIT = "draging out init";
local MODE_DRAGOUT = "draging out";

local REL_BACKGROUND = "background";
local REL_SELF = "self";
local REL_ABOVE = "above";
local REL_BELOW = "below";
local REL_ONTO = "onto";

local itemList = {};
local itemCount = 0;

---------------------------------------------------------------------------------------
local function setVisible(frame, visible)
	if visible then
		frame:Show();
	else
		frame:Hide();
	end
end

local function GetMenuGroup()
	return MAIN:GetFocus();
end

local function Visualize(self, item)
	self:SetHeight(HEIGHT_MAP[item.type] or HEIGHT_MAP["default"]);
	self.skeleton:SetPoint("LEFT", item.level and item.level * LEVEL_OFFSET or 0, 0);
	
	if item.type == EXTRA_TYPE["CATEGORY"] then
		self.bg:Show();
		self.icon:Show();
		self.icon:SetSize(SIZE_ICONSMALL, SIZE_ICONSMALL);
		self.icon:SetTexture(item.texture or item.data.texture);
		self.title:Show();
		self.title:SetText(item.text or item.data.name);
		self.subtitle:Show();
		self.subtitle:SetText(item.subtext or item.data.comment);
		self.texture:Hide();
		self.border:Hide();
	elseif item.type == EXTRA_TYPE["SEPARATOR"] then
		self.bg:Show();
		self.icon:Hide();
		self.title:Show();
		self.title:SetText("separator");
		self.subtitle:Hide();
		self.texture:Show();
		self.border:Hide();
	else
		self.bg:Show();
		self.icon:Show();
		self.icon:SetSize(SIZE_ICONLARGE, SIZE_ICONLARGE);
		self.icon:SetTexture(item.texture or item.data.texture);
		self.title:Show();
		self.title:SetText(item.text or item.data.name);
		self.subtitle:Show();
		self.subtitle:SetText(item.subtext or item.data.comment);
		self.texture:Hide();
		self.border:Show();
	end

	local enabled = not item.data.isGarbage and not item.data.outOfSync;
	local isDragged = ENTRIES.sourceIndex and ENTRIES.sourceIndex == self.index;
	if not enabled or isDragged then
		self.bg:SetVertexColor(unpack(COLOR_BG_DISABLED));
		self.title:SetTextColor(unpack(COLOR_FG_DISABLED));
		self.subtitle:SetTextColor(unpack(COLOR_FG_DISABLED));
		self.icon:SetDesaturated(true);
	elseif ENTRIES:GetFocus() == item.data then
		self.bg:SetVertexColor(unpack(COLOR_BG_HIGHLIGHTEN));
		self.title:SetTextColor(unpack(COLOR_FG_HIGHLIGHTEN));
		self.subtitle:SetTextColor(unpack(COLOR_FG_HIGHLIGHTEN));
		self.icon:SetDesaturated(false);
	else
		self.bg:SetVertexColor(unpack(COLOR_BG));
		self.title:SetTextColor(unpack(COLOR_FG));
		self.subtitle:SetTextColor(unpack(COLOR_FG));
		self.icon:SetDesaturated(false);
	end
	self:SetAlpha(isDragged and ALPHA_DRAGGED or 1.0);
	self:Show();
end

local function CreateRelatedTo(data, ...)
	local mg = GetMenuGroup();
	local created;
	if data then
		if data.type == EXTRA_TYPE["CATEGORY"] then
			created = mg:CreateSubentry(data, ...);
		else
			local parent = mg:ParentOf(data);
			local index = mg:IndexOf(data);
			created = mg:CreateSubentry(parent, index + 1, ...);
		end
	else
		created = mg:CreateSubentry(...);
	end
	if created then
		ENTRIES:BuildItems();
	end
	return created;
end

---------------------------------------------------------------------------------------	editor callbacks
local function macroEdit_OnAccept(...)
	CONFIG:SetLock(false);
	local data = ENTRIES:GetFocus();
	data.texture, data.name, data.cmd = ...;
	ENTRIES:Update();
end

local function textEdit_OnAccept(...)
	CONFIG:SetLock(false);
	local data = ENTRIES:GetFocus();
	data.name = ...;
	ENTRIES:Update();
end

local function editor_OnCancel()
	CONFIG:SetLock(false);
end

---------------------------------------------------------------------------------------	broker browser callbacks
local function BrokerSelector_OnAccepted(broker, motion)
	local created = CreateRelatedTo(ENTRIES:GetFocus(), EXTRA_TYPE["LAUNCHER"], broker.name, broker.icon, motion);
	ENTRIES:RefreshAndFocus(created);
	CONFIG:SetLock(false);
end

local function BrokerSelector_OnCancled()
	CONFIG:SetLock(false);
end

---------------------------------------------------------------------------------------	editors
local ItemEditor = {	--	editor for specific types
	[BASE_TYPE["MACRO"]] = function(parent, data, ...)
		MacroEditor:SetParent(parent);
		local name, texture, body = GetMacroInfo(data.data);
		MacroEditor:BeginEditing(data.type, texture or data.texture, name or data.name, body or data.cmd, nil, editor_OnCancel, ANCHOR_MANUAL, ...);
		MacroEditor:SetEditable(false);
	end,
	[EXTRA_TYPE["ADDONMACRO"]] = function(parent, data, ...)
		MacroEditor:SetParent(parent);
		MacroEditor:BeginEditing(data.type, data.texture, data.name, data.cmd, macroEdit_OnAccept, editor_OnCancel, ANCHOR_MANUAL, ...);
		MacroEditor:SetEditable(true);
	end,
	[EXTRA_TYPE["CATEGORY"]] = function(parent, data, ...)
		TextEditor:SetParent(parent);
		TextEditor:BeginEditing("Category Name", data.name, textEdit_OnAccept, editor_OnCancel, ANCHOR_MANUAL, ...);
	end,
};

---------------------------------------------------------------------------------------	convertor
local Convertor = {	--	convert specific types into ADDONMACRO
	[BASE_TYPE["COMPANION"]] = function(entry)
		entry.type = EXTRA_TYPE["ADDONMACRO"];
		return entry;
	end,
	[BASE_TYPE["BATTLEPET"]] = function(entry)
		entry.type = EXTRA_TYPE["ADDONMACRO"];
		entry.cmd = string.format("/run C_PetJournal.SummonPetByGUID(%s);", entry.data);
		return entry;
	end,
	[BASE_TYPE["SPELL"]] = function(entry)
		entry.type = EXTRA_TYPE["ADDONMACRO"];
		return entry;
	end,
	[BASE_TYPE["ITEM"]] = function(entry)
		entry.type = EXTRA_TYPE["ADDONMACRO"];
		return entry;
	end,
	[BASE_TYPE["MACRO"]] = function(entry)
		entry.type = EXTRA_TYPE["ADDONMACRO"];
		entry.name, entry.texture, entry.cmd = GetMacroInfo(entry.data);
		entry.char = nil;
		_addon:Sync_DecacheEntry(entry);
		return entry;
	end,
	[BASE_TYPE["EQUIPMENTSET"]] = function(entry)
		entry.type = EXTRA_TYPE["ADDONMACRO"];
		entry.name = entry.data;
		entry.texture = GetEquipmentSetInfoByName(entry.data);
		entry.cmd = string.format("/equipset %s\n", entry.data);
		entry.char = nil;
		return entry;
	end,
};

---------------------------------------------------------------------------------------	float btns
local function floatDragHandle_OnMouseDown(self, button)
	if ENTRIES.mode == MODE_NORMAL then
		ENTRIES:SetMode(MODE_DRAGOUT_INIT, self:GetParent():GetParent());
	end
end

local function floatDragHandle_OnMouseUp(self, button)
	if ENTRIES.mode == MODE_DRAGOUT_INIT then
		ENTRIES:SetMode(MODE_NORMAL);
	end
end

local function floatDragHandle_OnDragStart(self, button)
	if ENTRIES.mode == MODE_DRAGOUT_INIT then
		ENTRIES:SetMode(MODE_DRAGOUT, self);
	end
end

local function floatDragHandle_OnDragStop(self)
	if ENTRIES.mode ~= MODE_DRAGOUT then return; end
	ENTRIES:OnDragUpdate(0);	--	force update relation, relatedIndex
	local mg = GetMenuGroup();
	local relation = ENTRIES.relation;
	local relatedIndex = ENTRIES.relatedIndex;
	local dstItem = itemList[relatedIndex];
	local srcItem = itemList[ENTRIES.sourceIndex];
	
	if not relation then
		--	do nothing
	elseif relation == REL_BACKGROUND then
		--	onto background
		mg:Relocate(srcItem.data, mg.data);
	elseif relation == REL_SELF then
		--	do nothing
	elseif relation == REL_ONTO then
		mg:Relocate(srcItem.data, dstItem.data, 1);
	else
		local dstIndex;
		if srcItem.parent == dstItem.parent and dstItem.index > srcItem.index then
			dstIndex = relation == REL_ABOVE and dstItem.index - 1 or dstItem.index;
		else
			dstIndex = relation == REL_ABOVE and dstItem.index or dstItem.index + 1;
		end
		mg:Relocate(srcItem.data, dstItem.parent, dstIndex);
	end
	ENTRIES:BuildItems();
	ENTRIES:SetMode(MODE_NORMAL, srcItem.data);
end

local function floatMoveUp_OnClick(self, button)
	local index = self:GetParent():GetParent().index;
	local item = itemList[index];
	GetMenuGroup():Relocate(item.data, item.parent, item.index - 1);
	ENTRIES:BuildItems();
	ENTRIES:RefreshAndFocus(item.data);
end

local function floatMoveDown_OnClick(self, button)
	local index = self:GetParent():GetParent().index;
	local item = itemList[index];
	GetMenuGroup():Relocate(item.data, item.parent, item.index + 1);
	ENTRIES:BuildItems();
	ENTRIES:RefreshAndFocus(item.data);
end

local function floatEdit_OnClick(self, button)
	local parentBtn = self:GetParent():GetParent();
	local item = itemList[parentBtn.index];
	CONFIG:SetLock(true);
	ENTRIES:SetFocus(item.data);
	ItemEditor[item.type](ENTRIES, item.data, "TOPLEFT", parentBtn, 80, 0);
end

local function floatDelete_OnClick(self, button)
	local index = self:GetParent():GetParent().index;
	local data = itemList[index].data;
	data.isGarbage = true;
	if ENTRIES:GetFocus() == data then
		ENTRIES:SetFocus();
	else
		ENTRIES:Update();
	end
end

local function floatRecover_OnClick(self, button)
	local index = self:GetParent():GetParent().index;
	local item = itemList[index];
	item.data.isGarbage = nil;
	ENTRIES:Update();
end

local function HideFloatButtons()
	FloatButtons:Hide();
end

local function ShowFloatButtons(self, item)
	if not item then
		HideFloatButtons();
	else
		local isFocused = ENTRIES:GetFocus() == item.data;
		FloatButtons.enableColor = COLOR_FLOATBUTTON;
		FloatButtons.tipTextColor = isFocused and COLOR_FLOATBUTTON or COLOR_FLOATTIP;
		FloatButtons.disableColor = isFocused and COLOR_FLOATBUTTON_DISABLE_ALTER or COLOR_FLOATBUTTON_DISABLE;
		FloatButtons:RemoveButton(7, true);
		if item.data.isGarbage then
			FloatButtons:AddButton(MEDIA_PATH:format("round_refresh-32"), "recover", floatRecover_OnClick, 7, true);
		else
			FloatButtons:AddButton(MEDIA_PATH:format("round_delete-32"), "mark as garbage", floatDelete_OnClick, 7, true);
		end
		FloatButtons:GetButton(2):SetEnabled(not item.isTop);
		FloatButtons:GetButton(3):SetEnabled(not item.isBottom);
		FloatButtons:GetButton(5):SetEnabled(ItemEditor[item.type]);
		FloatButtons:AttachTo(self);
		FloatButtons:Update();
	end
end

---------------------------------------------------------------------------------------	scrollpane items
local function Button_OnEnter(self, motion)
	if ENTRIES.mode == MODE_NORMAL then
		for _, item in ipairs(ENTRIES.scrollPane.items) do
			item.highlight:Hide();
		end
		self.highlight:Show();
		ShowFloatButtons(self, itemList[self.index]);
	end
end

local function Button_OnLeave(self, motion)
	if motion and ENTRIES.mode == MODE_NORMAL then
		self.highlight:Hide();
		if self == FloatButtons:GetParent() then
			HideFloatButtons();
		end
	end
end

local function Button_OnMouseUp(self, button)
	if ENTRIES.mode == MODE_NORMAL then
		local item = itemList[self.index];
		ENTRIES:SetFocus(ENTRIES:GetFocus() ~= item.data and item.data);
		if self:IsMouseOver() then
			pcall(self:GetScript("OnEnter"), self);
		end
	elseif ENTRIES.mode == MODE_DROPIN then
		ENTRIES:OnDragUpdate(0);	--	force update relation, relatedIndex
		local dstItem = itemList[ENTRIES.relatedIndex];
		local relation = ENTRIES.relation;
		
		if not relation then
			--	do nothing
			ENTRIES:SetMode(MODE_NORMAL);
		elseif relation == REL_BACKGROUND then
			--	onto background
			local created = CreateRelatedTo(nil, GetCursorInfo());
			ENTRIES:SetMode(MODE_NORMAL, created);
		elseif relation == 	REL_SELF then
			--	wont happen
		elseif relation == REL_ONTO then
			local created = CreateRelatedTo(dstItem.data, 1, GetCursorInfo());
			ENTRIES:SetMode(MODE_NORMAL, created);
		else
			local created = CreateRelatedTo(dstItem.parent, relation == REL_ABOVE and dstItem.index or dstItem.index + 1, GetCursorInfo());
			ENTRIES:SetMode(MODE_NORMAL, created);
		end
	end
end

---------------------------------------------------------------------------------------
function ENTRIES:Init()
	do
		self.mode = MODE_NORMAL;
		self._elapsed = 0;
		self:SetTitle("Items", "Menu item manipulations");
		self:SetPrevious(MEDIA_PATH:format("round_left-32"), "Back to MenuGroups", function(self)
			CONFIG:SwitchPanel(CONFIG.main);
		end);
		self:SetScript("OnEvent", function(self, event, ...)
			local f = self[event];
			if f then
				f(self,...);
			end
		end);
		self.onActiveFunc = function(self)
			self:BuildItems();
			if not self:GetFocus() then
				self.scrollPane:SetScrollValue(0);
			else
				self:RefreshAndFocus(self:GetFocus());
			end

			self:RegisterEvent("ACTIONBAR_SHOWGRID");
			self:RegisterEvent("ACTIONBAR_HIDEGRID");
		end;
		self.onDeactiveFunc = function(self)
			self:SetMode(MODE_NORMAL);
			self:SetFocus(nil, true);
			HideFloatButtons();
			MacroEditor:CancelEditing();
			BrokerSelector:CancelEditing();
			self:UnregisterEvent("ACTIONBAR_SHOWGRID");
			self:UnregisterEvent("ACTIONBAR_HIDEGRID");
		end;
	end
	do	--	scrollPane
		local scrollPane = self.scrollPane;
		scrollPane.bar:SetControlable(true);
		scrollPane.bar:SetFadeAlpha(1, 0.5);
		scrollPane.GetOffsetItem = function(self, offsetPixel)
			for index, entry in ipairs(itemList) do
				local h = HEIGHT_MAP[entry.type] or HEIGHT_MAP["default"];
				offsetPixel = offsetPixel - h;
				if offsetPixel <= 0 then
					return index, offsetPixel + h;
				end
			end
		end;
		scrollPane.Update = function(self)
			self:GetParent():Update();
		end;

		self.scrollPane:SetScript("OnMouseUp", function(self)	--	MODE_DROPIN scrollpane receiver
			if ENTRIES.mode == MODE_DROPIN then
				local created = CreateRelatedTo(nil, GetCursorInfo());
				ENTRIES:SetMode(MODE_NORMAL, created);
			end
		end);

		self:RecreateTemplates();
	end
	do	--	charm buttons
		self.convertor:SetIcon(MEDIA_PATH:format("biotech-32"));
		self.convertor:SetText("Convert Macro+");
		self.convertor:SetScript("OnClick", function(self)
			local data = ENTRIES:GetFocus();
			data = Convertor[data.type](data);
			ENTRIES:BuildItems();
			ENTRIES:RefreshAndFocus(data);
		end);
		self.addFolder:SetIcon(MEDIA_PATH:format("add_folder-32"));
		self.addFolder:SetText("Add Folder");
		self.addFolder:SetScript("OnClick", function(self)
			local created = CreateRelatedTo(ENTRIES:GetFocus(), EXTRA_TYPE["CATEGORY"]);
			ENTRIES:RefreshAndFocus(created);
		end);
		self.addSeparator:SetIcon(MEDIA_PATH:format("border_color-32"));
		self.addSeparator:SetText("Add Separator");
		self.addSeparator:SetScript("OnClick", function(self)
			local created = CreateRelatedTo(ENTRIES:GetFocus(), EXTRA_TYPE["SEPARATOR"]);
			ENTRIES:RefreshAndFocus(created);
		end);
		self.addMacro:SetIcon(MEDIA_PATH:format("services-32"));
		self.addMacro:SetText("Add Macro+");
		self.addMacro:SetScript("OnClick", function(self)
			local created = CreateRelatedTo(ENTRIES:GetFocus(), EXTRA_TYPE["ADDONMACRO"]);
			ENTRIES:RefreshAndFocus(created);
		end);
		self.addBroker:SetIcon(MEDIA_PATH:format("usb_on-32"));
		self.addBroker:SetText("Add Broker");
		self.addBroker:SetScript("OnClick", function(self)
			CONFIG:SetLock(true);
			BrokerSelector:SetParent(ENTRIES);
			BrokerSelector:SetPreferedSize(4, 6);
			BrokerSelector:BeginEditing(
				GetMenuGroup():GetActiveBrokerNameList(), 
				BrokerSelector_OnAccepted, BrokerSelector_OnCancled, 
				ANCHOR_RELATIVE, self, 10, 5);
		end);
		self.gc:SetIcon(MEDIA_PATH:format("trash-32"));
		self.gc:SetText("Empty Garbages");
		self.gc:SetScript("OnClick", function(self)
			_addon:Config_GC(GetMenuGroup());
			ENTRIES:BuildItems();
		end);
	end
	do	--	assists
		GlassPane:SetScript("OnMouseUp", function(self, button)
			ENTRIES:SetMode(MODE_NORMAL);
		end);
	end
	do	--	FloatButtons
		FloatButtons:SetParent(CONFIG);
		local dragHandle = FloatButtons:AddButton(MEDIA_PATH:format("round_drag_32"), "drag to reorder", nil, 1, true);
		dragHandle:RegisterForDrag("LeftButton", "RightButton");
		dragHandle:SetScript("OnMouseDown", floatDragHandle_OnMouseDown);
		dragHandle:SetScript("OnMouseUp", floatDragHandle_OnMouseUp);
		dragHandle:SetScript("OnDragStart", floatDragHandle_OnDragStart);
		dragHandle:SetScript("OnDragStop", floatDragHandle_OnDragStop);
		FloatButtons:AddButton(MEDIA_PATH:format("round_up-32"), "move up", floatMoveUp_OnClick, 2, true);
		FloatButtons:AddButton(MEDIA_PATH:format("round_down-32"), "move down", floatMoveDown_OnClick, 3, true);
		FloatButtons:AddSeparator(4, true);
		FloatButtons:AddButton(MEDIA_PATH:format("round_edit-32"), "edit item", floatEdit_OnClick, 5, true);
		FloatButtons:AddSeparator(6, true);
		FloatButtons:AddButton(MEDIA_PATH:format("round_delete-32"), "mark as garbage", floatDelete_OnClick, 7, true);
	end
	do	--	virtual entries
		VirtualEntry.offsetX = 0;
		VirtualEntry.offsetY = 0;
		VirtualEntry.bg:SetVertexColor(unpack(COLOR_BG_HIGHLIGHTEN));
		VirtualEntry.text:SetTextColor(unpack(COLOR_FG_HIGHLIGHTEN));
		VirtualEntry:SetScript("OnShow", function(self)
			self.fadeIn:Play();
		end);
		VirtualEntry:SetScript("OnUpdate", function(self, elapsed)
			local scale, x, y = self:GetEffectiveScale(), GetCursorPosition();
			self:SetPoint("BOTTOMLEFT", nil,"BOTTOMLEFT", (x - self.offsetX) / scale, (y - self.offsetY) / scale);
		end);
		VirtualGapEntry.bg:SetVertexColor(unpack(COLOR_BG_HIGHLIGHTEN));
		VirtualGapEntry.text:SetTextColor(unpack(COLOR_FG_HIGHLIGHTEN));
		VirtualGapEntry:SetScript("OnShow", function(self)
			self.fadeIn:Play();
		end);
	end
end

function ENTRIES:Update()
	if not self:IsVisible() then return; end
	local scroller = self.scrollPane;
	setVisible(scroller.bgText, itemCount == 0);
	local offset = scroller:GetTopItemIndex();
	for i = 1, #scroller.items do
		local index = offset + i - 1;
		local button = scroller.items[i];
		button.index = index;
		if index <= itemCount then
			Visualize(button, itemList[index]);
			pcall(button:IsMouseOver() and 
				button:GetScript("OnEnter") or button:GetScript("OnLeave"), button);
		else
			button:Hide();
		end
	end
end

function ENTRIES:ScrollDataCenter(data)
	if not self:IsVisible() or not data then return; end
	for index, item in ipairs(itemList) do
		if item.data == data then
			local h = HEIGHT_MAP[item.type] or HEIGHT_MAP["default"];
			local offset = item.topOffset + h / 2;
			offset = offset + h / 2 - self.scrollPane:GetHeight() / 2;
			self.scrollPane:VerticalSmoothScrollTo(offset);
			return;
		end
	end
end

function ENTRIES:RefreshAndFocus(data)
	self:SetFocus(data);
	self:ScrollDataCenter(data);
end

function ENTRIES:RecreateTemplates()
	local items = self.scrollPane:Init("ActionMenuEntryTemplate", "Button");
	for _, btn in ipairs(items) do
		btn:RegisterForDrag("LeftButton");
		btn:SetScript("OnEnter", Button_OnEnter);
		btn:SetScript("OnLeave", Button_OnLeave);
		btn:SetScript("OnMouseUp", Button_OnMouseUp);
	end
end

function ENTRIES:_parseEntryList(parent, level, top)
	for index, data in ipairs(parent.subList) do
		local item = {
			level = level,
			parent = parent,
			index = index,
			data = data,
			type = data.type,

			subtext = STR_TYPES[data.type] or "Unknown",

			isBottom = index >= #parent.subList,
			isTop = index <= 1,
			topOffset = top,
		};
		tinsert(itemList, item);
		itemCount = itemCount + 1;
		top = top + (HEIGHT_MAP[data.type] or HEIGHT_MAP["default"]);

		if data.subList then
			item.subtext = string.format("%d item(s)", #data.subList);
			top = self:_parseEntryList(data, level + 1, top);
			item.bottomOffset = top;
		end
	end
	return top;
end

function ENTRIES:BuildItems()
	wipe(itemList);
	itemCount = 0;

	local totalHeight = self:_parseEntryList(GetMenuGroup().data, 0, 0);
	self.scrollPane:SetScrollRange(totalHeight);
end

function ENTRIES:SetFocus(data, noUpdate)
	if data and (data.isGarbage or data.outOfSync) then return; end
	self.focus = data;
	if not noUpdate and self:IsVisible() then
		self:Update();
	end
	self.convertor:SetEnabled(data and Convertor[data.type]);
end

function ENTRIES:GetFocus()
	return self.focus;
end

function ENTRIES:SetMode(mode, ...)
	self.mode = mode;

	for _, btn in ipairs(self.scrollPane.items) do
		btn.animDown:Stop();
	end
	if mode == MODE_NORMAL then
		local data = ...;
		self.relatedIndex = nil;
		self.relation = nil;
		self.sourceIndex = nil;
		self:SetScript("OnUpdate", nil);
		ClearCursor();
		GlassPane:Hide();
		VirtualEntry:Hide();
		VirtualGapEntry:Hide();
		self:RefreshAndFocus(data or self:GetFocus());
		FloatButtons:SetAlpha(1);
	elseif mode == MODE_DRAGOUT_INIT then
		local btn = ...;
		self.sourceIndex = btn.index;
		self:Update();
		GlassPane:Hide();
		local item = itemList[btn.index];
		local x, y = GetCursorPosition();
		VirtualEntry.icon:SetTexture(ICON_MOVE);
		VirtualEntry.icon:SetVertexColor(unpack(COLOR_FG_HIGHLIGHTEN));
		VirtualEntry.text:SetText(item.text or item.data.name);
		VirtualEntry:SetSize(btn:GetSize());
		VirtualEntry.offsetX = x - btn:GetLeft();
		VirtualEntry.offsetY = y - btn:GetBottom();
		VirtualEntry:ClearAllPoints();
		VirtualEntry:Hide();
		VirtualGapEntry.text:SetFormattedText("move <%s> here", item.text or item.data.name);
		VirtualGapEntry.icon:SetTexture(ICON_MOVE);
		VirtualGapEntry.icon:SetVertexColor(unpack(COLOR_FG_HIGHLIGHTEN));
		VirtualGapEntry:Hide();
	elseif mode == MODE_DRAGOUT then
		GlassPane:Show();
		VirtualEntry:Show();
		VirtualGapEntry:Hide();
		FloatButtons:SetAlpha(1e-9);
		--	onupdate go
		self._elapsed = 0;
		self:SetScript("OnUpdate", self.OnDragUpdate);
	elseif mode == MODE_DROPIN then
		local type = ...;
		GlassPane:Hide();
		VirtualEntry:Hide();
		VirtualGapEntry.text:SetFormattedText("Create <%s> entry", STR_TYPES[type] or "Unknown");
		VirtualGapEntry.icon:SetTexture(ICON_ADD);
		VirtualGapEntry.icon:SetVertexColor(unpack(COLOR_FG_HIGHLIGHTEN));
		VirtualGapEntry:Hide();
		--	onupdate go
		self._elapsed = 0;
		self:SetScript("OnUpdate", self.OnDragUpdate);
	end
end

function ENTRIES:OnDragUpdate(elapsed)
	self._elapsed = self._elapsed + elapsed;
	if THRESHOLD > self._elapsed then return end
	elapsed = mathFloor(self._elapsed / THRESHOLD) * THRESHOLD;
	self._elapsed = self._elapsed - elapsed;
	
	local btn;
	if self.scrollPane:IsMouseOver() then
		for _, b in ipairs(self.scrollPane.items) do
			if b:IsMouseOver() and b:IsShown() then
				btn = b;
				break;
			end
		end
	end
	
	local index, level, relation;
	local cursorX, cursorY = GetCursorPosition();
	
	if btn then
		index = btn.index;
		local item = itemList[index];
		level = item.level;
		local offset =  btn:GetTop() - cursorY;
		if index == self.sourceIndex then
			index = nil;
			relation = REL_SELF;
			level = nil;
		elseif offset > btn:GetHeight() * LOWERBOUND then
			if index + 1 == self.sourceIndex then
				index = nil;
				relation = REL_SELF;
				level = nil;
			else
				relation = REL_BELOW;
			end
		elseif offset < btn:GetHeight() * UPPERBOUND then
			if index - 1 == self.sourceIndex then
				index = nil;
				relation = REL_SELF;
				level = nil;
			else
				relation = REL_ABOVE;
			end
		else
			if item.type == EXTRA_TYPE["CATEGORY"] then
				relation = REL_ONTO;
				level = level + 1;
			else
				relation = nil;
			end
		end
	elseif self.scrollPane:IsMouseOver() then
		index = itemCount;
		relation = REL_BACKGROUND;
		level = 0;
	else
		-- mouse is not inbound
		index = nil;
		relation = nil;
		level = nil;
		-- auto scrolling
		local top, bottom = self.scrollPane:GetTop(), self.scrollPane:GetBottom();
		if cursorY > top then
			self.scrollPane:VerticalSmoothDeltaScroll(0.1, THRESHOLD);
		elseif cursorY < bottom then
			self.scrollPane:VerticalSmoothDeltaScroll(-0.1, THRESHOLD);
		end
	end

	-- if same situation as cached, do not react
	if self.relatedIndex == index and self.relation == relation then
		return;
	end
	-- visual effects
	local aboveIndex;
	if not relation then
		aboveIndex = nil;
	elseif relation == REL_ABOVE then
		aboveIndex = index - 1;
	else
		aboveIndex = index;
	end
	if aboveIndex then
		local target;
		for i, btn in ipairs(self.scrollPane.items) do
			btn.animDown:Stop();
			if btn.index == aboveIndex then
				target = btn;
			elseif btn.index > aboveIndex then
				btn.animDown:Play();
			end
		end
		VirtualGapEntry:ClearAllPoints();
		VirtualGapEntry:SetPoint("RIGHT", self.scrollPane);
		VirtualGapEntry:SetPoint("LEFT", self.scrollPane, "LEFT", level * LEVEL_OFFSET, 0);
		if target then
			VirtualGapEntry:SetPoint("TOP", target, "BOTTOM", 0, -5);
			VirtualGapEntry:Show();
		elseif aboveIndex == 0 then
			VirtualGapEntry:SetPoint("TOP", self.scrollPane);
			VirtualGapEntry:Show();
		else
			-- out of bound
			VirtualGapEntry:Hide();
		end
		VirtualEntry:Hide();
	else
		for _, btn in ipairs(self.scrollPane.items) do
			btn.animDown:Stop();
		end
		VirtualGapEntry:Hide();
		if self.mode == MODE_DRAGOUT then
			VirtualEntry:Show();
		end
	end
	-- cache
	self.relatedIndex = index;
	self.relation = relation;
end

---------------------------------------------------------------------------------------
function ENTRIES:ACTIONBAR_SHOWGRID()
	self:SetMode(MODE_DROPIN, GetCursorInfo());
end

function ENTRIES:ACTIONBAR_HIDEGRID()
	if self.mode ~= MODE_NORMAL then
		self:SetMode(MODE_NORMAL);
	end
end
