local _name, _addon = ...;

local CONFIG = Broker_ActionMenuConfig;
local MAIN = CONFIG.main;
local ENTRIES = CONFIG.entries;
local OPTIONS = CONFIG.options;
local FloatButtons = FloatButtons_Main;

local BASE_TYPE = _addon.BASE_TYPE;
local EXTRA_TYPE = _addon.EXTRA_TYPE;
local HEIGHT_CATEGORY = 35;
local HEIGHT_ENTRY = 40;
local SIZE_ICONSMALL = 24;
local SIZE_ICONLARGE = 32;

local CAT_ACCOUNT = "account";
local CAT_CHARACTER = "character";
local MEDIA_PATH = "Interface\\addons\\" .. _name .. "\\Medias\\%s";
local LEVEL_OFFSET = 15;

local COLOR_BG = {0.196, 0.196, 0.196};
local COLOR_BG_HIGHLIGHTEN = {1, 0.5, 0};
local COLOR_BG_DISABLED = {0.14, 0.14, 0.14};
local COLOR_FG = {1, 0.5, 0};
local COLOR_FG_HIGHLIGHTEN = {0.1, 0.1, 0.1};
local COLOR_FG_DISABLED = {0.5, 0.5, 0.5};
local COLOR_FLOATBUTTON = {0.17, 0.135, 0.118};
local COLOR_FLOATBUTTON_DISABLE = {0.337, 0.267, 0.231};
local COLOR_FLOATBUTTON_DISABLE_ALTER = {0.847, 0.451, 0};
local COLOR_FLOATTIP = {0.627, 0.588, 0.573};

local itemList = {};	--	index : item
local itemCount = 0;

---------------------------------------------------------------------------------------
local function setVisible(frame, visible)
	if visible then
		frame:Show();
	else
		frame:Hide();
	end
end

local function Visualize(self, item)
	local isPseudoCategory = item.level == 0;
	
	self.skeleton:SetPoint("LEFT", item.level and item.level * LEVEL_OFFSET or 0, 0);
	
	if isPseudoCategory then
		self:SetHeight(HEIGHT_CATEGORY);
		self.icon:SetSize(SIZE_ICONSMALL, SIZE_ICONSMALL);
		self.border:Hide();
	else
		self:SetHeight(HEIGHT_ENTRY);
		self.icon:SetSize(SIZE_ICONLARGE, SIZE_ICONLARGE);
		self.border:Show();
	end
	self.bg:Show();
	self.icon:Show();
	self.icon:SetTexture(item.texture or item.data.data.texture);
	self.title:Show();
	self.title:SetText(item.text or item.data.data.name);
	self.subtitle:Show();
	self.subtitle:SetText(item.subtext or item.data.data.comment);
	self.texture:Hide();

	local enabled = isPseudoCategory or not item.data:IsGarbage();
	if not enabled then
		self.bg:SetVertexColor(unpack(COLOR_BG_DISABLED));
		self.title:SetTextColor(unpack(COLOR_FG_DISABLED));
		self.subtitle:SetTextColor(unpack(COLOR_FG_DISABLED));
		self.icon:SetDesaturated(true);
	elseif MAIN:GetFocus() == item.data then
		self.bg:SetVertexColor(unpack(COLOR_BG_HIGHLIGHTEN));
		self.title:SetTextColor(unpack(COLOR_FG_HIGHLIGHTEN));
		self.subtitle:SetTextColor(unpack(COLOR_FG_HIGHLIGHTEN));
		self.icon:SetDesaturated(false);
	else
		self.bg:SetVertexColor(unpack(COLOR_BG));
		self.title:SetTextColor(unpack(COLOR_FG));
		self.subtitle:SetTextColor(unpack(COLOR_FG));
		self.icon:SetDesaturated(false);
	end
	self:SetEnabled(enabled);
	self:Show();
end

---------------------------------------------------------------------------------------	float btn funcs
local function charmNew_OnClick(self, button)
	local focused = MAIN:GetFocus();
	local menugroup;
	if type(focused) == "string" then
		menugroup = _addon:Core_CreateMenuGroup(focused);
	else
		local category, index = _addon:Core_IndexOfMenuGroup(focused);
		menugroup = _addon:Core_CreateMenuGroup(category, index + 1);
	end
	MAIN:BuildItems();
	MAIN:RefreshAndFocus(menugroup);
end

local function charmClone_OnClick(self, button)
	local menugroup = MAIN:GetFocus();
	menugroup = _addon:Core_CloneMenuGroup(menugroup);
	MAIN:BuildItems();
	MAIN:RefreshAndFocus(menugroup);
end

local function charmTrash_OnClick(self, button)
	_addon:Config_GC(MAIN:GetFocus());
	MAIN:BuildItems();
end

local function floatSwap_OnClick(self, button)
	local index = self:GetParent():GetParent().index;
	local menugroup = itemList[index].data;
	_addon:Core_SwapCategory(menugroup);
	MAIN:BuildItems();
	MAIN:RefreshAndFocus(menugroup);
end

local function floatEntries_OnClick(self, button)
	local index = self:GetParent():GetParent().index;
	local menugroup = itemList[index].data;
	MAIN:SetFocus(menugroup);
	CONFIG:SwitchPanel(ENTRIES);
end

local function floatOption_OnClick(self, button)
	local index = self:GetParent():GetParent().index;
	local menugroup = itemList[index].data;
	MAIN:SetFocus(menugroup);
	CONFIG:SwitchPanel(OPTIONS);
end

local function floatDelete_OnClick(self, button)
	local index = self:GetParent():GetParent().index;
	local menugroup = itemList[index].data;
	menugroup.data.isGarbage = true;
	if MAIN:GetFocus() == menugroup then
		MAIN:SetFocus();
	else
		MAIN:Update();
	end
end

local function floatRecover_OnClick(self, button)
	local index = self:GetParent():GetParent().index;
	local menugroup = itemList[index].data;
	menugroup.data.isGarbage = nil;
	MAIN:Update();
end

local function ShowFloatButtons(self, item)
	if FloatButtons._lastHovered and FloatButtons._lastHovered ~= self then
		pcall(FloatButtons._lastHovered:GetScript(FloatButtons._lastHovered:IsMouseOver() and "OnEnter" or "OnLeave"), FloatButtons._lastHovered);
		FloatButtons._lastHovered = nil;
	end
	if not item or item.level == 0 then
		FloatButtons:Hide();
	else
		local isFocused = MAIN:GetFocus() == item.data;
		FloatButtons.enableColor = COLOR_FLOATBUTTON;
		FloatButtons.tipTextColor = isFocused and COLOR_FLOATBUTTON or COLOR_FLOATTIP;
		FloatButtons.disableColor = isFocused and COLOR_FLOATBUTTON_DISABLE_ALTER or COLOR_FLOATBUTTON_DISABLE;
		FloatButtons:RemoveButton(6, true);
		FloatButtons:RemoveButton(1, true);
		if item.parent == CAT_CHARACTER then
			FloatButtons:AddButton(MEDIA_PATH:format("round_upload-32"), "move to account", floatSwap_OnClick, 1, true);
		else
			FloatButtons:AddButton(MEDIA_PATH:format("round_download-32"), "move to character", floatSwap_OnClick, 1, true);
		end
		local menugroup = item.data;
		if menugroup:IsGarbage() then
			FloatButtons:AddButton(MEDIA_PATH:format("round_refresh-32"), "recover", floatRecover_OnClick, 6, true);
		else
			FloatButtons:AddButton(MEDIA_PATH:format("round_delete-32"), "mark as garbage", floatDelete_OnClick, 6, true);
		end
		FloatButtons:AttachTo(self);
		FloatButtons:Update();
		FloatButtons._lastHovered = self;
	end
end

---------------------------------------------------------------------------------------
local ItemMeta = {
	["OnEnter"] = function(self, motion)
		local item = itemList[self.index];
		self.highlight:Show();
		ShowFloatButtons(self, item);
	end,
	["OnLeave"] = function(self, motion)
		if not self:IsMouseOver() then
			self.highlight:Hide();
			ShowFloatButtons(self);
		end
	end,
	["OnClick"] = function(self, button)
		local item = itemList[self.index];
		MAIN:SetFocus(MAIN:GetFocus() ~= item.data and item.data);
		if self:IsMouseOver() then
			pcall(self:GetScript("OnEnter"), self);
		end
	end,
};

---------------------------------------------------------------------------------------
function MAIN:Init()
	do
		self:SetTitle("Menu Groups", "Menu-group profiling");
		self:SetPrevious(MEDIA_PATH:format("round_left-32"), "Exit", function(self, button)
			CONFIG:Hide();
		end);
		self.onActiveFunc = function(self)
			self:BuildItems();
			local focused = self:GetFocus();
			if not focused then
				self.scrollPane:SetScrollValue(0);
			else
				self:RefreshAndFocus(focused);
			end
		end;
		self.onDeactiveFunc = function(self)
			FloatButtons:Hide();
		end;
	end
	do	--	scrollPane
		local scrollPane = self.scrollPane;
		scrollPane.bar:SetControlable(true);
		scrollPane.bar:SetFadeAlpha(1, 0.5);
		scrollPane.GetOffsetItem = function(self, offsetPixel)
			for index, _ in ipairs(itemList) do
				offsetPixel = offsetPixel - HEIGHT_ENTRY;
				if offsetPixel <= 0 then
					return index, offsetPixel + HEIGHT_ENTRY;
				end
			end
		end;
		scrollPane.Update = function(self)
			self:GetParent():Update();
		end;
		self:RecreateTemplates();
	end
	do	--	charm buttons
		self.new:SetIcon(MEDIA_PATH:format("add_file-32"));
		self.new:SetText("New");
		self.new:SetScript("OnClick", charmNew_OnClick);
		self.clone:SetIcon(MEDIA_PATH:format("copy-32"));
		self.clone:SetText("Clone");
		self.clone:SetScript("OnClick", charmClone_OnClick);
		self.gc:SetIcon(string.format(MEDIA_PATH, "trash-32"));
		self.gc:SetText("Collect Garbages");
		self.gc:SetScript("OnClick", charmTrash_OnClick);
	end
	do	--	FloatButtons
		FloatButtons:SetParent(CONFIG);
		FloatButtons:AddButton(MEDIA_PATH:format("round_upload-32"), "move to account", floatSwap_OnClick, 1, true);
		FloatButtons:AddSeparator(2, true);
		FloatButtons:AddButton(MEDIA_PATH:format("round_bullets_32"), "menugroup items", floatEntries_OnClick, 3, true);
		FloatButtons:AddButton(MEDIA_PATH:format("round_settings-32"), "menugroup options", floatOption_OnClick, 4, true);
		FloatButtons:AddSeparator(5, true);
		FloatButtons:AddButton(MEDIA_PATH:format("round_delete-32"), "mark as garbage", floatDelete_OnClick, 6, true);
		FloatButtons:Hide();
	end
end

function MAIN:Update()
	if not self:IsVisible() then return; end
	local scroller = self.scrollPane;
	setVisible(scroller.bgText, itemCount == 0);
	
	local offset = scroller:GetTopItemIndex();
	for i = 1, #scroller.items do
		local index = offset + i - 1;
		local button = scroller.items[i];
		button.index = index;
		if index <= itemCount then
			Visualize(button, itemList[index]);
			pcall(button:IsMouseOver() and 
				button:GetScript("OnEnter") or button:GetScript("OnLeave"), button);
		else
			button:Hide();
		end
	end
	
	if self._lastMouseoverButton and self._lastMouseoverButton:IsMouseOver() then
		pcall(self._lastMouseoverButton:GetScript("OnEnter"), self._lastMouseoverButton);
	end
end

function MAIN:ScrollDataCenter(menugroup)
	if not self:IsVisible() or not data then return; end
	for index, item in ipairs(itemList) do
		if item.data == data then
			local h = HEIGHT_MAP[item.type] or HEIGHT_MAP["default"];
			local offset = item.topOffset + h / 2;
			offset = offset + h / 2 - self.scrollPane:GetHeight() / 2;
			self.scrollPane:VerticalSmoothScrollTo(offset);
			return;
		end
	end
end

function MAIN:RefreshAndFocus(menugroup)
	if not self:IsVisible() then return; end
	self:SetFocus(menugroup);
	self:ScrollDataCenter(menugroup);
end

function MAIN:RecreateTemplates()
	local templates = self.scrollPane:Init("ActionMenuEntryTemplate", "Button");
	for _, template in ipairs(templates) do
		template:SetScript("OnEnter", ItemMeta.OnEnter);
		template:SetScript("OnLeave", ItemMeta.OnLeave);
		template:SetScript("OnClick", ItemMeta.OnClick);
	end
end

function MAIN:BuildItems()
	wipe(itemList);
	itemCount = 0;
	local totalHeight = 0;

	local count = _addon:Core_GetCategoryItemNumber(CAT_ACCOUNT);
	tinsert(itemList, {
		level = 0,
		texture = MEDIA_PATH:format("conference-32"),
		text = "Account",
		subtext = string.format("%d group(s)", count),
		data = CAT_ACCOUNT,
		
		topOffset = totalHeight,
	});
	itemCount = itemCount + 1;
	totalHeight = totalHeight + HEIGHT_ENTRY;
	for i = 1, count do
		local menugroup = _addon:Core_GetMenuGroupByIndex(CAT_ACCOUNT, i);
		local item = {
			level = 1,
			parent = CAT_ACCOUNT,
			index = i,
			data = menugroup,
			
			topOffset = totalHeight,
		};
		tinsert(itemList, item);
		itemCount = itemCount + 1;
		totalHeight = totalHeight + HEIGHT_ENTRY;
	end

	count = _addon:Core_GetCategoryItemNumber(CAT_CHARACTER);
	tinsert(itemList, {
		level = 0,
		texture = MEDIA_PATH:format("user-32"),
		text = string.format("Character : %s", GetUnitName("player", true)),
		subtext = string.format("%d group(s)", count),
		data = CAT_CHARACTER,
		
		topOffset = totalHeight,
	});
	itemCount = itemCount + 1;
	totalHeight = totalHeight + HEIGHT_ENTRY;
	for i = 1, count do
		local menugroup = _addon:Core_GetMenuGroupByIndex(CAT_CHARACTER, i);
		local item = {
			level = 1,
			parent = CAT_CHARACTER,
			index = i,
			data = menugroup,
			
			topOffset = totalHeight,
		};
		tinsert(itemList, item);
		itemCount = itemCount + 1;
		totalHeight = totalHeight + HEIGHT_ENTRY;
	end

	self.scrollPane:SetScrollRange(totalHeight);
end

function MAIN:SetFocus(focus, noUpdate)	--	focus here is menugroup (or pseudo category)
	self.focus = focus;
	if not noUpdate and self:IsVisible() then
		self:Update();
	end
	local isMenuGroup = focus and type(focus) == "table";
	self.new:SetEnabled(focus);
	self.clone:SetEnabled(isMenuGroup);
end

function MAIN:GetFocus()
	return self.focus;
end

