local _name, _addon = ...;

local CONFIG = Broker_ActionMenuConfig;
local MAIN = CONFIG.main;
local ENTRIES = CONFIG.entries;
local OPTIONS = CONFIG.options;

local MEDIA_PATH = "Interface\\addons\\" .. _name .. "\\Medias\\%s";
local DEFAULTICON = "Interface\\Icons\\INV_Misc_QuestionMark";

---------------------------------------------------------------------------------------
local function GetMenuGroup()
	return MAIN:GetFocus();
end

---------------------------------------------------------------------------------------
function OPTIONS:Init()
	do
		self:SetTitle("Options", "Menu advanced configurations");
		self:SetPrevious(MEDIA_PATH:format("round_left-32"), "Back to MenuGroups", function(self)
			CONFIG:SwitchPanel(CONFIG.main);
		end);
		self.onActiveFunc = function(self)
			local data = GetMenuGroup().data;
			self:Reload();
			self.scrollPane:VerticalSmoothScrollTo(0);	--todo
		end;
		self.onDeactiveFunc = function(self)
			self:ClearFocus();
		end;
	end
	do	-- scrollpane
		local content = self.scrollPane;
		content.bar:SetControlable(true);
		content.bar:SetFadeAlpha(1, 0.5);
		do
			local item = CreateFrame("Button", "group1title", content, "ScrollSeparatorTemplate");
			item:Init("Generic");
			content:StackPush(item);
			item = CreateFrame("Button", "MenuGroupName", content, "ScrollTextfieldInputTemplate");
			item:Init("Name", "", function(text)
				if text and strlen(text) > 0 then
					local mg = GetMenuGroup();
					mg.data.name = text;
				else
					return true;
				end
			end);
			item:RegisterCallback(self, "Item_OnFocusChanged");
			content:StackPush(item);
			content.name = item;
			
			item = CreateFrame("Button", "MenuGroupComment", content, "ScrollTextfieldInputTemplate");
			item:Init("Comment", "", function(text)
				local mg = GetMenuGroup();
				mg.data.comment = text;
			end);
			item:RegisterCallback(self, "Item_OnFocusChanged");
			content:StackPush(item);
			content.comment = item;
		end
		do
			local item = CreateFrame("Button", "group1title", content, "ScrollSeparatorTemplate");
			item:Init("LibDataBroker related");
			content:StackPush(item);

			item = CreateFrame("Button", "MenuGroupIcon", content, "ScrollIconSelectorTemplate");
			item:Init("Icon", DEFAULTICON, function(texture)
				local mg = GetMenuGroup();
				mg.data.texture = texture;
				mg:UpdateBrokerObj();
			end);
			item.iconSelector:SetParent(CONFIG);	--	in order to bypass container lock
			item:RegisterCallback(self, "Item_OnFocusChanged");
			content:StackPush(item);
			content.icon = item;
			
			item = CreateFrame("Button", "MenuGroupLabel", content, "ScrollTextfieldInputTemplate");
			item:Init("Label", "", function(text)
				local mg = GetMenuGroup();
				mg.data.label = text;
				mg:UpdateBrokerObj();
			end);
			item:RegisterCallback(self, "Item_OnFocusChanged");
			content:StackPush(item);
			content.label = item;
			
			item = CreateFrame("Button", "MenuGroupText", content, "ScrollTextfieldInputTemplate");
			item:Init("Text", "", function(text)
				local mg = GetMenuGroup();
				mg.data.text = text;
				mg:UpdateBrokerObj();
			end);
			item:RegisterCallback(self, "Item_OnFocusChanged");
			content:StackPush(item);
			content.text = item;
		end
		do
			local item = CreateFrame("Button", "group2title", content, "ScrollSeparatorTemplate");
			item:Init("Menu tweaks");
			content:StackPush(item);
			
			item = CreateFrame("CheckButton", "MenuGroup_showFolderIcon", content, "ScrollTogglerTemplate");
			item:SetTextures(string.format(MEDIA_PATH, "checked_checkbox-32"), string.format(MEDIA_PATH, "unchecked_checkbox-32"));
			item:Init("Use Category Icon", false, function(state)
				local mg = GetMenuGroup();
				mg.data.showFolderIcon = state;
			end);
			item:RegisterCallback(self, "Item_OnFocusChanged");
			content:StackPush(item);
			content.showFolderIcon = item;
			
			item = CreateFrame("CheckButton", "MenuGroup_persisted", content, "ScrollTogglerTemplate");
			item:SetTextures(string.format(MEDIA_PATH, "checked_checkbox-32"), string.format(MEDIA_PATH, "unchecked_checkbox-32"));
			item:Init("Always show menu", false, function(state)
				local mg = GetMenuGroup();
				mg.data.persisted = state;
			end);
			item:RegisterCallback(self, "Item_OnFocusChanged");
			content:StackPush(item);
			content.persisted = item;
			
			item = CreateFrame("CheckButton", "MenuGroup_hideMissing", content, "ScrollTogglerTemplate");
			item:SetTextures(string.format(MEDIA_PATH, "checked_checkbox-32"), string.format(MEDIA_PATH, "unchecked_checkbox-32"));
			item:Init("Hide missing spell/item", false, function(state)
				local mg = GetMenuGroup();
				mg.data.hideMissing = state;
			end);
			item:RegisterCallback(self, "Item_OnFocusChanged");
			content:StackPush(item);
			content.hideMissing = item;
			
			item = CreateFrame("CheckButton", "MenuGroup_hideCooling", content, "ScrollTogglerTemplate");
			item:SetTextures(string.format(MEDIA_PATH, "checked_checkbox-32"), string.format(MEDIA_PATH, "unchecked_checkbox-32"));
			item:Init("Hide cooldown spell/item", false, function(state)
				local mg = GetMenuGroup();
				mg.data.hideCooling = state;
			end);
			item:RegisterCallback(self, "Item_OnFocusChanged");
			content:StackPush(item);
			content.hideCooling = item;
		end
	end
end

function OPTIONS:Reload()
	local content = self.scrollPane;
	local data = GetMenuGroup().data;
	
	content.name:SetText(data.name);
	content.comment:SetText(data.comment);
	
	content.icon:SetIcon(data.texture);
	content.label:SetText(data.label);
	content.text:SetText(data.text);
	
	content.persisted:SetChecked(data.persisted);
	content.showFolderIcon:SetChecked(data.showFolderIcon);
	content.hideMissing:SetChecked(data.hideMissing);
	content.hideCooling:SetChecked(data.hideCooling);
	
	for _, item in ipairs(self.scrollPane.stack) do
		pcall(item:GetScript(item:IsMouseOver() and "OnEnter" or "OnLeave"), item);
	end
end

function OPTIONS:GetFocus()
	return self.focus;
end

function OPTIONS:ClearFocus(exception)
	for _, item in ipairs(self.scrollPane.stack) do
		if item ~= exception then
			item:SetFocus(false);
		end
	end
	self.focus = exception;
end

function OPTIONS:Item_OnFocusChanged(item, focus)
	if item == self.scrollPane.icon then
		CONFIG:SetLock(focus);
	end
	if focus then
		self:ClearFocus(item);
	end
end
